#define ISIZE 16
typedef unsigned short INT;

// return true if specified bit is set, 0 = LSB
#define bit(a, pos) (a[(pos)>>4]&(1<<((pos)&15)))

// return number of highest set bit, 256 if highest, 1 if lowest, 0 if none
static int msb(const INT *a)
{
  int i;

  for (i=ISIZE; i; i--)
    if (a[i-1])
      for(i*=16; i; i--)
        if (bit(a,i-1))
          return(i);
  return(0);
}

// return sum of 'digits' of a
static long sum(const INT *a)
{
  int i; long l;
  for (i=ISIZE, l=0; i--;)l+=*a++;
  return(l);
}

// increment a by n, return 1 if carry
static int addn(INT *a, unsigned short n)
{
  int i;

  if ((*a+=n) >= n) return(0);
  for (i=ISIZE; --i;) if (++*(++a)) return(0);
  return(1);
}

// a+=b, return overflow
static int add(INT *a, const INT *b)
{
  int i;
  unsigned long c = 0;

  for (i=ISIZE; i--;) *a++ = (c = (long)*a + *b++ + (c>0xffff));
  return (c>0xffff);
}

// decrement a by short n, return 1 if borrow
static int subn(INT *a, unsigned short n)
{
  int i;

  if ((*a-=n) <= ~n) return(0);
  for (i=ISIZE; --i;) if (~(--*(++a))) return(0);
  return(1);
}

// a-=b, return underflow
static int sub(INT *a, const INT *b)
{
  int i;
  unsigned long c = 0;

  for (i=ISIZE; i--;) *a++ = (c = ((long)*a - (c>0xffff)) - *b++);
  return (c>0xffff);
}


// a=n
static void movn(INT *a, unsigned short n)
{
  int i;

  *a++=n;
  for (i=ISIZE; --i;) *a++=0;
}

// a=b
static void mov(INT *a, const INT *b)
{
  int i;

  for(i=ISIZE; i--;) *a++=*b++;
}


// compare a to n, 0 if a = n, 1 if a>n, -1 if a<n
static int cmpn(const INT *a, unsigned short n)
{
  int i;

  if (*a != n)
  {
    if (*a < n) return(-1);
    return(1);
  }
  for (i=ISIZE; --i;) if (*++a) return(1);
  return(0);
}

// compare a to b (unsigned), 0 if a=b, 1 if a>b, -1 if a<b
static int cmp(const INT *a, const INT *b)
{
  int i;

  for(i=ISIZE; i--;) if (a[i] != b[i])
  {
    if (a[i] < b[i]) return(-1);
    return(1);
  }
  return 0;
}

// a<<=1, return carry
static int shl(INT *a, int carry)
{
  unsigned short x;
  int i, c;

  for(i=ISIZE; i--;)
  {
    x = *a;
    c = x & 0x8000;
    x <<= 1;
    if (carry) x |= 1;
    carry = c;
    *a++=x;
  }
  return(carry);
}


// a/=b, r=remainder (a%b)
static void div(INT *a, const INT *b, INT *r)
{
  int i;                              // number of bits

  movn(r,0);
  for (i=ISIZE*16; i--;)
  {
    shl(r,shl(a,0));                  // shl a into mod
    if (cmp(r,b)>=0)                  // if r >= b
    {
      sub(r,b);                       // mod it
      *a |= 1;                        // note it
    }
  }
}

// a*=b
static void mul(INT *a, const INT *b)
{
  int i, h;
  INT c[ISIZE];

  if (!(h = msb(b)))
  {
    movn(a,0);
    return;
  }

  mov(c,a);                     // c = b
  movn(a,0);                    // A = 0

  for(i=0; i < h; i++)          // for each bit
  {
    if (bit(b,i)) add(a, c);    // a+=c if bit d[i] is set
    shl(c,0);                   // c <<= 1
  }
}

// a=(a*b)%mod
static void mulmod(INT *a, const INT *b, const INT *mod)
{
  int i, h;
  INT c[ISIZE];

  if (!(h = msb(b)))
  {
    movn(a,0);
    return;
  }

  div(a,mod,c);                // set c to initial modulus
  movn(a,0);                   // c = 0

  for(i=0; i < h; i++)         // for each bit
  {
    if (bit(b,i))
    {
      add(a,c);
      if (cmp(a,mod)>=0) sub(a,mod);
    }
    shl(c,0);
    if (cmp(c,mod)>=0) sub(c,mod); // if c >= mod, c-=mod
  }
}

// cube with mod
static void cubemod(INT *a, const INT *mod)
{
  INT t[ISIZE];

  mov(t,a);
  mulmod(a,t,mod);
  mulmod(a,t,mod);
}

// square with mod
static void sqrmod(INT *a, const INT *mod)
{
  INT t[ISIZE];
  mov(t,a);
  mulmod(a,t,mod);
}

// a=a^b%mod
static void expmod(INT *a, const INT *b, const INT *mod)
{
  int i, h;
  INT acc[ISIZE];

  if (!(h = msb(b)))                    // if no highest bit
  {
    movn(a,1);                          // we're done
    return;
  }

  movn(acc,1);
  for (i=0; i<h; i++)                   // for each bit in b
  {
    if (bit(b,i)) mulmod(acc,a,mod);    // if set, accumlate
    sqrmod(a,mod);                      // anyway, square a
  }
  mov(a,acc);
}



// a = greatest common divisor(a,b)
static void gcd(INT *a, const INT *b)
{
  INT r[ISIZE], bb[ISIZE];
  mov(bb,b);

  while (cmpn(bb,0))                    // while b != 0
  {
    div(a,bb,r);                        // r = a%b
    mov(a,bb);                          // a = b
    mov(bb,r);                          // b = r
  }
}

// a=(1/a) mod b
// i.e. what number, multiplied with a, modulo b, is 1?
// use extended euclidian algorithm, woo hoo
// maybe optimizable for a=3?
static void invmod(INT *a, const INT *b)
{
  INT rp[ISIZE], rc[ISIZE], rn[ISIZE],
      tc[ISIZE], tp[ISIZE], qc[ISIZE];

  mov(rp,b);
  mov(rc,a);
  movn(tc,1);
  movn(tp,0);

  for(;;)
  {
    mov(qc,rp); div(qc,rc,rn);   // qc=rp/rc, rn=rp%rc
    mulmod(qc,tc,b);             // tn = (tp-tc*qc) % b
    if (sub(tp,qc)) add(tp,b);
    div(tp,b,a);

    switch(cmpn(rn,1))           // check remainder
    {
      case -1:                   // if 0
        movn(a,0);               // force error

      case 0:                    // if 1
        return;                  // result is in a
    }

    mov(tp,tc);                  // tp=tc
    mov(tc,a);                   // tc=a
    mov(rp,rc);                  // rp=rc
    mov(rc,rn);                  // rc=rn
  }
}

// find next prime > a
// prime if 2^(a-1) mod a = 1
// not only slow, but damn slow
void next_prime(INT *a)
{
  INT b[ISIZE], c[ISIZE];

  *a |= 1;                      // ensure oddness
  do
  {
    addn(a,1);                  // a+=2
    mov(b,a); subn(b,1);        // b=a-1
    movn(c,2); expmod(c,b,a);   // c=2^b mod a
  } while (cmpn(c,1));          // until c=1
}

// generate decrypt and encrypt keys assuming seed = 3
// d and e should be initially randomized
void genkeys(INT *d, INT *e)
{
  INT pq[ISIZE];

  do next_prime(d);             // get next prime d
  while (!((sum(d)-1)%3));      // while d-1 is a multiple of 3

  do next_prime(e);             // same for e
  while (!((sum(e)-1)%3));      // yeah, magic, but it works!

  mov(pq,e);
  mul(e,d);
  subn(pq,1);
  subn(d,1);
  mul(pq,d);

  movn(d,3);                    // d = inverse of 3 mod (p-1)(q-1)
  invmod(d,pq);
}

#include <stdio.h>

// set a to random number of bits
void rnd(INT *a, int bits)
{
  int x, p = 0;

  movn(a,0);

  while(bits--)
  {
    x = rand();
    if (x >= p) a[bits>>4] |= 1<<(bits&15);
    p = x;
  }
}

// convert INT to string with leading zeros suppressed
#define nibble(a, n) (((a[n>>2])&(15<<((n&3)<<2)))>>((n&3)<<2))

static char * INT2s(const INT *a)
{
  static char b[(ISIZE*4)+1];
  char *s = b;
  int i, n;

  if (!cmpn(a,0)) *s = '0';
  else
  {
    for (i = ISIZE*4; !nibble(a,i-1); i--);
    for(;i;i--)
    {
      n = nibble(a,i-1);
      if (n <= 9) *s++ = n+'0';
      else *s++ = (n-10)+'A';
    }
  }
  *s = 0;
  return(b);
}

// convert string to INT
static void s2INT(INT *a, const char *s)
{
  int i;

  movn(a,0);

  for(;;)
  {
    if (*s >= '0' && *s <= '9') i = *s-'0';
    else if (*s >= 'A' && *s <= 'F') i = (*s-'A')+10;
    else break;
    shl(a,0);
    shl(a,0);
    shl(a,0);
    shl(a,0);
    *a |= i;
  }
}

// encrypt M E, writes C

void main(int argc, char *argv[])
{

  INT D[ISIZE], E[ISIZE], M[ISIZE], K[ISIZE];

  if (argc != 2)
  {
    printf("Usage: encrypt E");
    return;
  }

  s2INT(E,argc[1]);
  bits = msb(E);



  rnd(E,msb(M));
  rnd(D,msb(M));

  genkeys(D,E);                     // convert to 'real' keys

  // encrypt message M with key E
  cubemod(M,E);

  // decrypt message M with keys D and E
  expmod(M,D,E);
}
