#include "crypto.h"

#define BSIZE 16  // allow 128 bit keys
typedef unsigned short BIGNUM[BSIZE];

// return true if specified bit is set, 0 = LSB
#define bit(a, pos) (a[(pos)>>4]&(1<<((pos)&15)))

// return number of highest set bit, 256 if highest, 1 if lowest, 0 if none
static int msb(BIGNUM a)
{
  int i;

  for (i=BSIZE; i; i--)
    if (a[i-1])
      for(i*=16; i; i--)
        if (bit(a,i-1))
          return(i);
  return(0);
}

// increment a by n, return 1 if carry
static int addn(BIGNUM a, unsigned short n)
{
  int i;

  if ((*a+=n) >= n) return(0);
  for (i=BSIZE; --i;) if (++*(++a)) return(0);
  return(1);
}

// a+=b, return overflow
static int add(BIGNUM a, BIGNUM b)
{
  int i;
  unsigned long c = 0;

  for (i=BSIZE; i--;) *a++ = (c = (unsigned long)*a + *b++ + (c>0xffff));
  return (c>0xffff);
}

// decrement a by short n, return 1 if borrow
static int subn(BIGNUM a, unsigned short n)
{
  int i;

  if ((*a-=n) <= ~n) return(0);
  for (i=BSIZE; --i;) if (~(--*(++a))) return(0);
  return(1);
}

// a-=b, return underflow
static int sub(BIGNUM a, BIGNUM b)
{
  int i;
  unsigned long c = 0;

  for (i=BSIZE; i--;) *a++ = (c = ((unsigned long)*a - (c>0xffff)) - *b++);
  return (c>0xffff);
}


// a=n
static void movn(BIGNUM a, unsigned short n)
{
  int i;

  *a++=n;
  for (i=BSIZE; --i;) *a++=0;
}

// a=b
static void mov(BIGNUM a, BIGNUM b)
{
  int i;

  for(i=BSIZE; i--;) *a++=*b++;
}


// compare a to n, 0 if a = n, 1 if a>n, -1 if a<n
static int cmpn(BIGNUM a, unsigned short n)
{
  int i;

  if (*a != n)
  {
    if (*a < n) return(-1);
    return(1);
  }
  for (i=BSIZE; --i;) if (*++a) return(1);
  return(0);
}

// compare a to b (unsigned), 0 if a=b, 1 if a>b, -1 if a<b
static int cmp(BIGNUM a, BIGNUM b)
{
  int i;

  for(i=BSIZE; i--;) if (a[i] != b[i])
  {
    if (a[i] < b[i]) return(-1);
    return(1);
  }
  return 0;
}

// a<<=1, return carry
static int shl(BIGNUM a, int carry)
{
  unsigned short x;
  int i, c;

  for(i=BSIZE; i--;)
  {
    x = *a;
    c = x & 0x8000;
    x <<= 1;
    if (carry) x |= 1;
    carry = c;
    *a++=x;
  }
  return(carry);
}

static int shr(BIGNUM a, int carry)
{
  unsigned short x;
  int i, c;

  for(i=BSIZE; i--;)
  {
    x = a[i];
    c = x & 1;
    x >>=1;
    if (carry) x |= 0x8000;
    carry = c;
    a[i]=x;
  }
  return(carry);
}

// a/=b, r=remainder (a mod b)
static void div(BIGNUM a, BIGNUM b, BIGNUM r)
{
  int i;                              // number of bits

  movn(r,0);
  for (i=BSIZE*16; i--;)
  {
    shl(r,shl(a,0));                  // shl a into remainder
    if (cmp(r,b)>=0)                  // if r >= b
    {
      sub(r,b);                       // keep in range
      *a |= 1;                        // remember as quotient
    }
  }
}

// a*=b
static void mul(BIGNUM a, BIGNUM b)
{
  int i, h;
  BIGNUM c;

  if (!(h = msb(b)))
  {
    movn(a,0);
    return;
  }

  mov(c,a);                     // c = b
  movn(a,0);                    // A = 0

  for(i=0; i < h; i++)          // for each bit
  {
    if (bit(b,i)) add(a, c);    // a+=c if bit d[i] is set
    shl(c,0);                   // c <<= 1
  }
}

// a=(a*b) mod m
static void mulmod(BIGNUM a, BIGNUM b, BIGNUM m)
{
  int i, h;
  BIGNUM c;

  if (!(h = msb(b)))
  {
    movn(a,0);
    return;
  }

  div(a,m,c);                           // c = a mod m
  movn(a,0);

  for(i=0; i < h; i++)                  // for each bit
  {
    if (bit(b,i))                       // if set
    {
      add(a,c);                         // accumulate
      if (cmp(a,m)>=0) sub(a,m);        // keep a in range
    }
    shl(c,0);                           // c*=2
    if (cmp(c,m)>=0) sub(c,m);          // also keep in range
  }
}

// a=a^b mod m
static void expmod(BIGNUM a, BIGNUM b, BIGNUM m)
{
  int i, h;
  BIGNUM acc, t;

  if (!(h = msb(b)))                    // if no highest bit
  {
    movn(a,1);                          // we're done
    return;
  }

  movn(acc,1);
  for (i=0; i<h; i++)                   // for each bit in b
  {
    if (bit(b,i)) mulmod(acc,a,m);      // if set, accumlate
    mov(t,a); mulmod(a,t,m);            // anyway, square a
  }
  mov(a,acc);
}

// Convert BIGNUM to KEY, encoded in base 95 using characters '!' to '~'
// return 0 if result exceeds size (including 0 termination)
static void ntok(BIGNUM a, KEY k)
{
  BIGNUM t, r;

  movn(t,95);
  while(cmpn(a,0))
  {
    div(a,t,r);
    *k++=*r+' ';
  }
  *k=0;
}

// Convert base 95 encoded KEY to BIGNUM
// stops at first illegal character
static void kton(KEY k, BIGNUM a)
{
  BIGNUM t;
  char *p = k;

  while(*p >= ' ' && *p <= '~') p++;

  movn(t,95);
  movn(a,0);

  while (p-->k)
  {
    mul(a,t);
    addn(a,*p-' ');
  }
}

#ifdef ENCRYPT
// encrypt message in place
void encrypt(KEY E, KEY message)
{
  BIGNUM m,e,t;

  kton(message,m);              // convert strings to nums
  kton(E,e);

  mov(t,m);                      // m = m^3 mod e
  mulmod(m,t,e);
  mulmod(m,t,e);

  ntok(m,message);
}
#endif

#ifdef DECRYPT
// decrypt message in place, return 0 if didn't fit
void decrypt(KEY D, KEY E, KEY message)
{
  BIGNUM m, d, e;
  kton(message,m);
  kton(D,d);
  kton(E,e);
  expmod(m,d,e);
  ntok(m,message);
}
#endif

#ifdef KEYGEN
// a=(1/a) mod b
// i.e. what number, multiplied with a, modulo b, is 1?
// use extended euclidian algorithm
static void invmod(BIGNUM a, BIGNUM b)
{
  BIGNUM rp, rc, rn, tc, tp, qc;

  mov(rp,b);
  mov(rc,a);
  movn(tc,1);
  movn(tp,0);

  for(;;)
  {
    mov(qc,rp); div(qc,rc,rn);   // qc=rp/rc, rn=rp mod rc
    mulmod(qc,tc,b);             // tn = (tp-tc*qc) mod b
    if (sub(tp,qc)) add(tp,b);
    div(tp,b,a);

    switch(cmpn(rn,1))           // check remainder
    {
      case -1:                   // if 0
        movn(a,0);               // force error

      case 0:                    // if 1
        return;                  // result is in a
    }

    mov(tp,tc);                  // tp=tc
    mov(tc,a);                   // tc=a
    mov(rp,rc);                  // rp=rc
    mov(rc,rn);                  // rc=rn
  }
}

// find next prime > a
// prime if 2^(a-1) mod a = 1
// not only slow, but damn slow
void next_prime(BIGNUM a)
{
  BIGNUM b, c;

  *a |= 1;                      // ensure oddness
  do
  {
    addn(a,1);                  // a+=2
    mov(b,a); subn(b,1);        // b=a-1
    movn(c,2); expmod(c,b,a);   // c=2^b mod a
  } while (cmpn(c,1));          // until c=1
}

// return sum of 'digits' of a
static long sum(BIGNUM a)
{
  int i; long l;
  for (i=BSIZE, l=0; i--;)l+=*a++;
  return(l);
}

// generate keys E and D from initially random strings
// key length is limited to specified number of bits
// (init strings should be longer than this...)
int keygen(KEY D, KEY E, int bits)
{
  BIGNUM e, d, t;
  int x;

  kton(D,d);                    // make 'random' number
  kton(E,e);
  mov(t,d);

  while(msb(d)<bits) mul(d,e);
  for(x=msb(d); x>(bits/2); x--) shr(d,0);

  while(msb(e)<bits) mul(e,t);
  for(x=msb(e); x>(bits/2); x--) shr(e,0);

  do next_prime(d);             // get next prime d
  while (!((sum(d)-1)%3));      // while d-1 is a multiple of 3

  do next_prime(e);
  while (!((sum(e)-1)%3));

  mov(t,e);
  mul(e,d);                     // e=e*d
  subn(t,1);
  subn(d,1);
  mul(t,d);                     // t = e-1*d-1

  movn(d,3);                    // d = inverse of 3 mod (e-1)(d-1)
  invmod(d,t);

  x = msb(e);
  ntok(e,E);
  ntok(d,D);
  return(x);
}

#endif
