#!/usr/bin/perl
#
# This script will use HTTP to pull down the most recent news in XML
# format.  A small position file is maintained so that we can request only
# the news we haven't yet seen.  See the usage text (run "pullxml.pl -h")
# for how to specify where you want the output to go (e.g. into batch files
# in a user-specified dir or to STDOUT).
#
my $VERSION = '4.04 (May 22, 2003)';

use strict;

use Getopt::Long;
use IO::Socket;

my $DEFAULT_CGI_HOST = 'feedcenter.clari.net';
my $DEFAULT_CGI_URL = '/perl/pullfeed';

# The technical details of the interaction with the web server is simple:
#
# 1. We send a POST request to the host and path mentioned above.  The data
# items we send are as follows:  host, ipos, limit, and plain.  Set the
# "host" to the pull-host you registered with ClariNet.  Set "ipos" to
# "1,0" as a default, and thereafter set it to the "cur" value from the
# <position cur="1234,5678" end="2345,6789"> tag.  (This value will match
# the "end" value when there is no more news to fetch.)  Set "limit" to an
# approximate batch size you want to receive.  Set "plain" to "1" so that
# the CGI returns plain text data and not HMTL-encoded text.

my($all_flag, $output_dir, $help_opt, $start_pos, $tmpfile_in_parent_dir);
my($version_opt, $verbosity);

my $cgi_host = $DEFAULT_CGI_HOST;
my $cgi_url = $DEFAULT_CGI_URL;
my $feed_fqdn = '';
my $limit = 256*1024;
my $saw_clarifeed_tag = 0;
my $exit_value = 0;

&Getopt::Long::Configure('bundling');
&usage if !&GetOptions(
    'all|a' => \$all_flag,
    'cgi-host|c=s' => \$cgi_host,
    'output-dir|d=s' => \$output_dir,
    'feed-name|f=s' => \$feed_fqdn,
    'help|h' => \$help_opt,
    'pull-limit|l=i' => \$limit,
    'start-pos|p=s' => \$start_pos,
    'parent-temp|P' => \$tmpfile_in_parent_dir,
    'cgi-url|u=s' => \$cgi_url,
    'version' => \$version_opt,
    'verbose|v+' => \$verbosity,
) || $help_opt;

die "Version: $VERSION\n" if $version_opt;

&usage if @ARGV != 1;

my $logpos_file = shift;

my($log_pos, $end_pos, $tmpfile);
local *OUT;

if (defined($output_dir)) {
    chdir($output_dir) or die "Unable to chdir to $output_dir\n";
    $tmpfile = ($tmpfile_in_parent_dir? '../' : '') . "pull.$$";
}

if (open(LOG_POS, "+<$logpos_file")) {
    chomp($log_pos = <LOG_POS>);
}
else {
    if (!open(LOG_POS, "+>$logpos_file")) {
	die "Unable to create $logpos_file\n";
    }
}
if (defined($start_pos)) {
    ($log_pos) = $start_pos =~ /^(\d+,\d+)$/;
    die "Invalid value for -p\n" if !defined($log_pos);
}
elsif (!defined($log_pos)) {
    $log_pos = "1,0";
}

while (1) {
    my $pullfp = IO::Socket::INET->new(Proto => "tcp", PeerAddr => $cgi_host,
				    PeerPort => "http(80)" );
    die "cannot connect to HTTP daemon on $cgi_host" unless $pullfp;
    $pullfp->autoflush(1);
    my $data = "host=$feed_fqdn&ipos=$log_pos&limit=$limit&plain=1";
    my $len = length($data);
    print STDERR "POST data: $data\n" if $verbosity >= 2;
    print $pullfp <<EOT;
POST $cgi_url HTTP/1.0
Content-Length: $len

$data
EOT

    my $outfp;
    if (defined($output_dir)) {
	open($outfp = *OUT, ">$tmpfile") or die "Unable to write $output_dir/$tmpfile";
    }
    else {
	$outfp = *STDOUT;
    }

    while (<$pullfp>) {
	s/\r$//;
	print STDERR $_ if $verbosity >= 4;
	last if /^$/;
    }
    while (<$pullfp>) {
	if (/^<(article|image|cancel) /) {
	    my $end_match = qr#^</$1>$#;
	    print $outfp $_;
	    while (<$pullfp>) {
		print $outfp $_;
		last if /$end_match/;
	    }
	}
	elsif (/^<position cur="(.*?)" end="(.*?)"/) {
	    $log_pos = $1;
	    $end_pos = $2;
	    print STDERR $_ if $verbosity;
	}
	elsif (/^<\?xml / || /^<!DOCTYPE /) {
	    print $outfp $_ unless $saw_clarifeed_tag;
	    print STDERR $_ if $verbosity >= 2;
	}
	elsif (/^<clarifeed>/) {
	    print $outfp $_ unless $saw_clarifeed_tag;
	    print STDERR $_ if $verbosity >= 2;
	    $saw_clarifeed_tag = 1;
	}
	elsif (m#^</clarifeed>#) {
	    print STDERR $_ if $verbosity >= 2;
	}
	elsif (/^<errormsg>/) {
	    print STDERR $_ if $verbosity >= 2;
	    $_ = <$pullfp>;
	    print STDERR $_;
	    print $outfp $_;
	    $_ = <$pullfp>;
	    print STDERR $_ if $verbosity >= 2;
	}
	elsif (/^#! ERROR (.+)/) {
	    print STDERR $_;
	    print $outfp $_;
	}
	else {
	    print STDERR "TRASH: $_" if $verbosity >= 4;
	}
    }
    close($pullfp);

    if ($end_pos) {
	seek(LOG_POS, 0, 0);
	print LOG_POS $log_pos, "\n";
	truncate(LOG_POS, tell(LOG_POS));
    }
    if (defined($output_dir)) {
	if ($saw_clarifeed_tag) {
	    print $outfp "</clarifeed>\n";
	    $saw_clarifeed_tag = 0;
	}
	close($outfp);
	if (-s $tmpfile) {
	    my $cnt = 0;
	    while (!link($tmpfile, time . $cnt++ . '.t')) {
		; # Check the error?
	    }
	}
	unlink($tmpfile);
    }
    if (!$end_pos) {
	$exit_value = 1;
	last;
    }

    last if $log_pos eq $end_pos;
    if (!defined($output_dir) && !$all_flag) {
	$exit_value = 2;
	last;
    }
    sleep(10);
    undef $end_pos;
}

print "</clarifeed>\n" if $saw_clarifeed_tag;

exit $exit_value;


sub usage
{
    my($fp) = $help_opt? *STDOUT : *STDERR;
    print $fp <<EOT;
Usage: pullxml.pl [-OPTS] LOGPOS_FILE

-a, --all             In STDOUT mode (the default), instead of only outputting
                      the result of a single pull, pull as many times as needed
                      to catch up, outputting all the pulls on STDOUT.
-c, --cgi-host=HOST   Specify the ClariNet HOST from which we pull the news.
-d, --output-dir=DIR  Output unique-named batch files into the specified
                      DIR instead of sending the data to STDOUT.  The finished
                      filenames will begin with a number (e.g. 123456789.t),
                      so make sure you don't grab the tmp file (e.g. pull.123).
-f, --feed-name=HOST  Specify the HOST that identifies our pull feed.
-h, --help            Display this help message.
-l, --pull-limit=N    Each pull will end when the total size goes past N bytes.
-p, --start-pos=P     Set the starting logfile position ("1,0" is today, "2,0"
                      is yesterday).  Only use this for reprocessing.
-P, --parent-temp     Use output-dir's parent dir to write out the tmp file.
-u, --cgi-url=PATH    Override the URL we use to pull the news.
-v, --verbose         Set the verbosity level (repeat to increase).
--version             Output the version and exit. [Ver: $VERSION]

For example:

    perl pullxml.pl -f my.feed.name /path/position.file >news.xml
    perl pullxml.pl -f my.feed.name -d /some/spool/dir pos
EOT
    exit 1;
}
