#!/usr/bin/perl -w
my $version = "XX0204";
print "LSL compiler $version (refrigerate after opening)\n";

use strict;
$|=1;

my $debug = 0;        # set default debug level
my $file;
my $oldstyle = 0;

sub choke { die "Usage: lsl [-dN] [-m] filename\n" }
# parse cmd line args
for (@ARGV)
{
  /^-d(\d+)$/i and $debug = $1, next;
  /^-m$/i and $oldstyle = 1, next;
  /^-/ and choke;
  $file and choke;
  $file = $_;
}  
  
choke unless $file;

-f $file && open F,$file or die "Can't open $file\n";
my @source = <F>;
close F;

# ROM specific variables 
sub MAXRAM() {4096}                         # maximum svar ram allocation

sub MAXFILE() {16384-128-1}                 # maximum file size

sub OPCODES() { 
 [['glue',0x01,                             # glue n, svar, allow, row, col, width, max, up, dn, lt, rt
    qw(1-16 s S? 1-6 1-79 1-79? 1-190? 1-16? 1-16? 1-16? 1-16?)], 
  ['verify',0x02,qw(s 1-190 t)],            # verify svar, min, target
  ['set',0x03,qw(s S?)],                    # set svar, string
  ['set',0x04,qw(n N?)],                    # set nvar, number
  ['set',0x05,qw(s s 1-190? 1-190?)],       # set svar, svar, start, length 
  ['set',0x06,qw(n n?)],                    # set nvar, nvar (args required)
  ['set',0x07,qw(s n)],                     # set svar, nvar (convert)
  ['set',0x08,qw(n s)],                     # set nvar, svar (convert)
  ['eq',0x09,qw(s S t)],                    # eq svar,string,target
  ['eq',0x0a,qw(n N t)],                    # eq nvar,number,target
  ['eq',0x0b,qw(s s t)],                    # eq svar,svar,target
  ['eq',0x0c,qw(n n t)],                    # eq nvar,nvar,target
  ['ne',0x0d,qw(s S t)],                    # ne svar,string,target
  ['ne',0x0e,qw(n N t)],                    # ne nvar,number,target
  ['ne',0x0f,qw(s s t)],                    # ne svar,svar,target
  ['ne',0x10,qw(N N t)],                    # ne number,number,target
  ['lt',0x11,qw(s s t)],                    # lt svar,string,target
  ['lt',0x12,qw(n N t)],                    # lt nvar,number,target
  ['lt',0x13,qw(s s t)],                    # lt svar,svar,target
  ['lt',0x14,qw(n n t)],                    # lt nvar,nvar,target
  ['gt',0x15,qw(s S t)],                    # gt svar,string,target
  ['gt',0x16,qw(n N t)],                    # gt nvar,number,target
  ['gt',0x17,qw(s s t)],                    # gt svar,svar,target
  ['gt',0x18,qw(n n t)],                    # gt nvar,nvar,target
  ['add',0x19,qw(n n)],                     # add nvar,nvar
  ['add',0x1a,qw(n N)],                     # add nvar,number
  ['mul',0x1b,qw(n n)],                     # mul nvar,nvar
  ['mul',0x1c,qw(n N)],                     # mul nvar,number
  ['div',0x1d,qw(n n)],                     # div nvar,nvar
  ['div',0x1e,qw(n N)],                     # div nvar,number
  ['sub',0x1f,qw(n n)],                     # sub nvar,nvar
  ['sub',0x20,qw(n N)],                     # sub nvar,number
  ['mod',0x21,qw(n n)],                     # mod nvar,nvar
  ['mod',0x22,qw(n N)],                     # mod nvar,number
  ['cat',0x25,qw(s S)],                     # cat svar,string
  ['cat',0x26,qw(s s)],                     # cat svar,svar
  ['goto',0x27,qw(t)],                      # goto target
  ['return',0x28],                          # return 
  ['pop',0x29],                             # pop
  ['display',0x2a,qw(S)],                   # display inline_string
  ['display',0x2a,qw(.)],                   # display here_string 
  ['put',0x2b,qw(s 1-6 1-79? 1-79?)],       # put svar, line, col, width
  ['put',0x2c,qw(S 1-6 1-79? 1-79?)],       # put inline string, line, col, width
  ['put',0x2c,qw(. 1-6 1-79? 1-79?)],       # put here string, line, col, width
  ['softkey',0x2d,qw(1-8 S8 t?)],           # softkey n, string, target
  ['csoftkey',0x2e,qw(1-8 S8 t?)],          # csoftkey n, string, target
  ['leftkey',0x2f,qw(1-6 t?)],              # leftkey n, target
  ['rightkey',0x30,qw(1-6 t?)],             # rightkey n, target
  ['idle',0x31,qw(N t?)],                   # idle timeout, target
  ['wait',0x39,qw(0-0! t?)],                # wait 0,target (magic)  
  ['wait',0x32,qw(N? t?)],                  # wait timeout, target
  ['connect',0x33,qw(t? t? t? t? t? 0-1?)], # connect ok, busy, noconnect, dropper, server, xsession
  ['capture',0x34],                         # capture  
  ['alias',0x35,qw(0-9 s)],                 # alias n, svar
  ['count',0x36,qw(n s S?)],                # count nvar, svar, "string"
  ['pos',0x37,qw(n s s)],                   # pos nvar, svar, svar
  ['tag',0x38,qw(s? S?)]]};
  
sub MAXSVARS() {320};                       # max defineable svars

sub SVARS() {                               # predefined svars
   {SERVER     => 36480,                    # the actual one and only server connect message
    VERSION    => 36479,                    # rom version 
    SCRIPT     => 36478,                    # script ID 
    TITLE      => 36477,                    # title
    URL        => 36476,                    # passed to lsp server
    ALIAS0     => 36475,                    # aliases, set by 'alias' command
    ALIAS1     => 36474,
    ALIAS2     => 36473,
    ALIAS3     => 36472,
    ALIAS4     => 36471,
    ALIAS5     => 36470,
    ALIAS6     => 36469,
    ALIAS7     => 36468,
    ALIAS8     => 36467,
    ALIAS9     => 36466,
    NUMBER     => 36465,
    ID         => 36464,                    # device id
    CHAT       => 36463,                    # chat script
    SAVE0      => 36462,                    # generic script parameters
    SAVE1      => 36461,                  
    SAVE2      => 36460,                  
    SAVE3      => 36459,                  
    SAVE4      => 36458,                  
    SAVE5      => 36457,                  
    SAVE6      => 36456,
    SAVE7      => 36455,
    STAMP      => 36454}}
    
sub MAXNVARS() {64};                        # max definable nvars

sub NVARS() {                               # predefined nvars                           
   {FREE       => 190,                      # free bytes remaining
    USED       => 189,                      # bytes used by data files
    RECORDS    => 188,                      # number of records in memory 
    CALLBACK   => 187,                      # desired callback seconds from lsp server 
    LED        => 186,                      # led mode
    TIMER0     => 185,                      # generic timers
    TIMER1     => 184,          
    TIMER2     => 183,          
    TIMER3     => 182,          
    TIMER4     => 181,          
    TIMER5     => 180,          
    TIMER6     => 179,          
    TIMER7     => 178,          
    IDLE       => 177,                      # seconds since last key hit
    BACKLIGHT  => 176,                      # seconds until backlight off (when > 60) 
    LINE       => 175,                      # seconds since line activity 
    SCONFIG    => 174,                      # serial configuration
    SERIAL     => 173,                     
    SAVAIL     => 172,                     
    SPEND      => 171,                     
    RANDOM     => 170,                     
    CLOCK      => 169,
    CONVERT    => 168}}
                                
sub SS() {3}                                # bytes in string index

sub NN() {4}                                # bytes in a numbr

sub SV() {2}                                # bytes in an svar index 

sub NV() {1}                                # bytes in an nvar index

sub TT() {2}                                # bytes in a target

sub STRFILES {4}                            # number of string files

sub MAXNUM() {(191**4)-1}                   # max integer value

# lots of good debug info
sub debug(@)
{
  if ($debug > 2)
  {
    my $depth=1;
    my @callers;
    unshift @callers,(caller $depth++)[3] while caller $depth;
    print "[@{[map ((split /::/)[1],@callers)]}] ",map(defined $_?"'$_' ":"undef ",@_),"\n";
  }  
}  

# after this, die will print source line number...
my $line = 0;
$SIG{__DIE__} = sub { print STDERR "line $line: " };

sub within($$$)
{
  my ($x, $min, $max) = @_;
}  

sub BASE() {191}

# convert positive number to base 191 encoded string of specified length 
# assumes zero if undefined
# we'll assume no more than 4 digits are ever required
sub encode($$)
{
  my $n = shift || 0;
  my $digits = shift or die "no digits?";
  my @n;

  die "$n isn't numeric\n" unless $n =~ /^\d+/;

  $n = $n % (MAXNUM+1);
  
  while ($n)
  {
    unshift @n,$n%BASE;
    $n = int $n/BASE;
  }
 
  unshift @n,0 while @n < $digits;
  
  grep $_+=(($_>94)?65:32), @n; 
 
  pack "c*",@n;
}  

# eval a number, put it in range of 0-MAXNUM
sub number($$)
{
  debug @_;
  my $r = undef;
  my $t = shift || 0;
  my $test = shift;
  local $SIG{__WARN__}= sub{};
  $t =~ s/\$(.)/ord $1/eg;
  eval "\$r=$t" unless $t =~ /'"/;
  return defined $r if $test;
  die "bad numeric expression\n" unless defined $r;
  $r % (MAXNUM+1); # -1 % MAXNUM+1 is MAXNUM, etc.
}

my $CODE = "";   # generated code 
# compile literal (hopefully encoded) strings
sub compile($)
{
  die "script ID not defined\n" unless $CODE;
  debug @_;
  $CODE .= $_[0];
  die "code segment size exceeds ",MAXFILE," bytes\n" if length $CODE > MAXFILE;
}

### stuff to handle symbolic targets (for goto, etc)
my %forward;          # forward referenced targets
my %target;         # resolved targets

# given target name, set forward reference if unknown and return 0
# else return address
sub target($)
{
  my $s = shift;
  return 0 unless $s;
  return $target{$s} if $target{$s};             # return known value 
  push @{$forward{$s}},length $CODE;             # or make forward references
  debug "forward $s -> @{$forward{$s}}"; 
  0;                                             # compile a dummy 
}  

# define target at current code offset
# resolve forward references, if required
sub maketarget($)
{
  my $s = shift;
  die "missing target\n" unless $s;
  die "duplicate target $s\n" if $target{$s};
  $target{$s}=length $CODE;                     # remember code address 
   
  debug "$s=$target{$s}";
  if ($forward{$s})                             # any forward references?
  {
    for (@{$forward{$s}})
    {
      debug "resolve $_";
      substr $CODE,$_,TT,encode($target{$s},TT); # resolve them
    }  
    delete $forward{$s};                        # note resolved
  }
}


# strings
my @STRS; 
{
  my %strs;           # string addresses 
  # add string to table if unknown
  # then return address or 'null' pointer
  sub EOT (){"\r"}    # string table field separator (carriage return)
  sub string($)
  {
    my $s = shift;
    return 0 unless $s;
    unless ($strs{$s})         # already defined?
    {
      my $l = length $s;       # no, get length
      for my $n (0..STRFILES-1)
      {
        die "script ID not defined\n" unless $STRS[$n];
        my $sl = length $STRS[$n];
        if ($sl+$l < MAXFILE)
        {
          $strs{$s}=($n*16384)+$sl;
          $STRS[$n] .= $s . EOT;
          goto OK;
        }
      }
      die "string segment size exceeds ",STRFILES*MAXFILE," bytes\n";
    }
    OK:
    debug "$s -> $strs{$s}";
    $strs{$s};
  }  
}


# string variables
my $VARS="";     # output section
my $ram = 0;     # ram consumed by all svars
my %svars = ();  # user defined string variables
my %nvars = ();  # user defined numeric variables

# return index of named svar
sub issvar($)
{
  my $s = shift;

  ($s && ($svars{$s} || ${&SVARS}{$s})) || 0;
}  
    
# return index of named svar or error
sub svar($)
{
  my $s = shift;

  return 0 unless defined $s;

  issvar $s or die "$s svar undefined\n";
}

# return index of named nvar
sub isnvar($)
{
  my $s = shift;

  ($s && ($nvars{$s} || ${&NVARS}{$s})) || 0;
}

# return index of named nvar or error
sub nvar($)
{
  my $s = shift;
   
  isnvar $s or die "$s nvar undefined\n";
} 

# create svar reference
sub makesvar(@)
{
  die "script ID not defined\n" unless $VARS;

  my ($name, $max)=@_; 
  die "$name duplicate variable\n" if issvar($name) || isnvar($name);
  die "svar name invalid" if /^\d/; 
  $svars{$name} = (scalar keys %svars)+1; 
  die "too many svars\n" if keys %svars > MAXSVARS;
  $max or $max = 190; 
  $ram += $max + 1;
  die "combined svar length exceeds ",MAXRAM," bytes\n" if $ram > MAXRAM;
  
  $VARS .= encode($max,1);
  die "var file size exceeds ",MAXFILE," bytes\n" if length $VARS > MAXFILE;
} 

# create nvar reference
sub makenvar(@)
{
  my $name = shift;
  die "$name duplicate variable\n" if issvar($name) || isnvar($name);
  die "nvar name invalid" if /^\d/; 
  $nvars{$name} = (scalar keys %nvars)+1; 
  die "too many nvars\n" if keys %nvars > MAXNVARS;
}

my $scriptid = "";
sub scriptid($)
{
  $scriptid = shift;
  die "redefinition of script id\n" if $CODE;
  
  $CODE = $scriptid . "\n";  # put in code section
  $VARS = $scriptid . "\n";  # and vars section
  $STRS[$_] = $scriptid . "\n" for 0..STRFILES-1; 
}  
  
# parser
my @arg;  # list of input line arguments
my @str;  # corresponding list indicating string arguments

sub yank()
{
  my $max = 6;
  my @s;     
  {
    $line++;
    die "expected '.'\n" unless @source; 
    $_=shift @source; 
    s/\s+$//;
    s/\s/ /g;
    last if /^\.$/;
    push @s,$_;
    die "too many lines\n" if @s > $max;
    redo;
  }
  my $s = join "\n",@s;
  $s =~ s/\\(\\|")/$1/g;  # convert \\ and \";
  debug $s;
  $s;
}
     
# main loop starts here
LINE: while (@source)
{
  $line++;
  $_ = shift @source;
  # first, break the line into arg[] list
  s/\s/ /g;
  my @line = split //;
  @arg=();
  @str=();

  TOKEN: while (@line)
  {
    my $c = shift @line;
    my $t;

    $c eq ' ' and next;
    $c eq "#" and last TOKEN;
    $c eq "," and push (@str,0), push(@arg,undef), next;
    $c eq "\"" and do
    {
      $t = '';
      while (@line)
      {
        my $c = shift @line;
        $c eq "\"" and push (@str,1), push(@arg,$t), goto ABSORB;
        $t .= ($c eq "\\")?shift(@line):$c;
      }
      die "missing close quote\n";
    }; 
    
    $t = $c;
    push (@str,0);
    while (@line)
    {
      my $c = shift @line;
      $c eq "#" and last;
      $c eq "," and push(@arg,$t), next TOKEN;
      $c eq " " and push(@arg,$t), goto ABSORB; 
      $t .= $c;
    } 
    push @arg,$t;
    last TOKEN;
  
    ABSORB: 
    next TOKEN if @arg == 1;
    while (@line)
    { 
      my $c = shift @line;
      $c eq " " and next;
      $c eq "#" and last TOKEN;
      $c eq "," and next TOKEN;
      die "expected a comma\n";
    }
  }

  # now process parsed tokens
  $arg[0] || next;

  if ($debug)
  {
    printf "L$line V%.4X C%.4X",length $VARS, length $CODE;
    $STRS[$_-1] && printf " S$_%.4X", length $STRS[$_-1] for 1..STRFILES;
    print ": ";
    print defined $arg[$_]?($str[$_]?"\"$arg[$_]\" ":"$arg[$_] "):"- " for 0..$#arg;  
    print "\n";
  }  
     
  # handle opcodes that don't actually compile anything
  $arg[0] =~ /^id$/ && do                  # script id
  {
    die "missing operand\n" unless defined $arg[1];
    scriptid $arg[1];
    next LINE;
  };
    
  # string name,"tag","allowed",min,max
  $arg[0] =~ /^string$/ && do
  {
    die "missing name\n" unless defined $arg[1];
    die "too many operands\n" if @arg > 3;
    makesvar $arg[1],$arg[2];
    next LINE;
  };
 
  # number name  alias for 'var name,,,2
  $arg[0] =~ /^number$/ && do
  {
    die "missing name\n" unless defined $arg[1];
    makenvar $arg[1];
    next LINE;
  };

  # :target
  $arg[0] =~ /:(.+)/ && do
  {
    maketarget $1;
    next LINE;
  }; 

  my @op; # operand list
  my $found = 0;
  # do version specific stuff
  SEEK: {
    OP: for (@{OPCODES()})                 # for each opcode
    {
      @op = @{$_};
      next if $op[0] ne $arg[0];        # if name match
      $found = 1;                       # remember
      next OP if @arg > $#op;           # next if too many given args
      for my $n (1..$#op-1)             # for each defined operand
      {
        my $o = $op[$n+1];             
        if ($o =~ /^\./)                # if should be a 'here-string' 
        {
          next OP if defined $arg[$n];  # next if arg defined
          next;                         # else ok
        } 
        if (defined $arg[$n])           # if arg defined
        {
          next OP if $o =~ /\./;        # next if supposed to be a 'here-string'
          next OP if $o =~ /^S/ && !$str[$n];                 # or if must be string but isn't
          next OP if $o =~ /^S(\d+)/ && length $arg[$n] > $1; # or string too long
          next OP if $o =~ /^n/ && !isnvar $arg[$n];          # or undefined nvar
          next OP if $o =~ /^s/ && !issvar $arg[$n];          # or undefined svar
          next OP if $o =~ /^N/ && !number($arg[$n],1);       # or not a number (expression)
          next OP if $o =~ /^(\d+)-(\d+)/ &&                  # or value out of range
            ($arg[$n] =~ /\D/ || $arg[$n] < $1 || $arg[$n] > $2); 
        } else 
        {
          next OP if $o !~ /\?$/;       # else next if arg wasn't optional
        }  
      }
      last SEEK;
    }  
    die $found?"invalid operand\n":"invalid opcode\n";
  }  
  
  # ok, compile @op
  compile encode($op[1],1);
  for (1..$#op-1)
  {
    my $o = $op[$_+1];
    $o =~ /!$/ and next; # trailing ! indicates lexical only
    $o =~ /^s/ and compile encode(svar $arg[$_],SV), next;  
    $o =~ /^n/ and compile encode(nvar $arg[$_],NV), next;
    $o =~ /^S/ and compile encode(string $arg[$_],SS), next;
    $o =~ /^N/ and compile encode(number($arg[$_],0),NN), next;
    $o =~ /^t/ and compile encode(target $arg[$_],TT), next;
    $o =~ /^(\d+)-(\d+)/ and compile encode($arg[$_],1), next;
    $o =~ /^\./ and compile encode(string yank,SS), next;
    die "lsl compiler bug! Invalid operand type $o\n";
  }  
}     

die "unresolved forward references: @{[keys %forward]}\n" if keys %forward;

# done, create output files
my $stub = "$scriptid.$version.".(STRFILES+2);

# cod=255, str1=254, str2=253, var=252, str3=251, str4=250

my @out = (\$CODE, \$STRS[0], \$STRS[1], \$VARS);
push @out, (\$STRS[2], \$STRS[3]) if STRFILES == 4;

for (0..$#out)
{
  my $f;
  if ($oldstyle)
  {
    if ($_== 0) { $f = "lslcod.out" }
    elsif ($_ == 1) { $f = "lslstr1.out" }
    elsif ($_ == 2) { $f = "lslstr2.out" }
    elsif ($_ == 3) { $f = "lslvar.out" }
    else { $f = "lslstr".($_-1).".out" }
  } else
  {
    $f = "lslout.".(scalar @out).".".(255-$_);
  }  
  print "$f: ";
  open F,">$f" or die " error $!\n";
  binmode F;
  print F ${$out[$_]};
  print "OK\n";
  close F;
}

# statistics
print scalar keys %svars," svars defined, allocating $ram bytes of ram\n";  
print scalar keys %nvars," nvars defined\n";  

sub blocks($) {((shift)+0x1FF)>>9}

my $cod = 1+length $CODE;
my $var = 1+length $VARS;

print "code size           : ",blocks $cod," blocks ($cod bytes)\n",
      "var size            : ",blocks $var," blocks ($var bytes)\n";
        
my $str = 0;
for (1..STRFILES)
{
  my $l = 1+length $STRS[$_-1];
  $str += $l;
  print "string $_ size       : ", blocks $l," blocks ($l bytes)\n";
}  

