# Mailbug Browser
package mbrowse;

use LWP::UserAgent;
use URI;
use HTML::TokeParser;
use HTML::Entities;
use lsp;
use warnings;
use strict;

# home url, default is command line but changeable with magic <sethome href="x"> tag
my $home_page;

# trueif current page is trusted, ie the home page or localhost or www.landel.com 
my $trusted=0; 

# page title
my $title;

my $baseref;

# stack of visited urls as [$url, $top, $selected], i.e. for back. $stack[0]->[0] is 'home'
my @stack;

sub base()
{
  my $b = $baseref || $stack[$#stack]->[0];
  $b =~ s/\?.*//;
  $b;
}

# page rendered text, in format ($text, $align) 
my @lines;

# this array of user selectable items, in format (line, col, \@element)
my @items;

# this is an array of form elements or links 
# link           -> (0, $url)
# input text     -> (1, /@forms name, value (editable), maxlength, size)
# input password -> (2, /@forms name, value (editable), maxlength, size) 
# input checkbox -> (3, /@forms name, value, checked)
# input hidden   -> (4, /@forms name, value, checked)
# input submit   -> (5, /@forms name, value)
my @elements;

# this is an array of forms, in format ($url, \@element...\@element)
my @forms;

# hash of submit \@element references, keyed by line 
my %leftkeys;
my %rightkeys;
my %softkeys;

# given html source, build the page arrays above 
{ # local variables for rendering routines
  my $parser; my $attr;                                 # for TokeParser                           
  my $align = 0;                                        # alignment: 0=left, 1=center, 2=pleft, 4=pcenter, 8=pright
  my $pre = 0;                                          # 1 = text is preformatted
  
sub blank() { ["",0] };               # a blank line

sub break() { (@lines > 1 && $lines[$#lines-1]->[0] =~ /\S/) || $lines[$#lines]->[0] =~ /\S/ and push @lines, blank; };  # push blank line
sub cbreak() { $lines[$#lines]->[0] =~ /\S/ and (push @lines, blank), return; $lines[$#lines]->[0] = ''}

sub nbsp() { chr 160 } 
 
sub magic() { chr 139 } # character used to indicate links, part of special mailbug character set 

# get next html token via HTML::TokeParser
sub token()
{
  my $t = $parser->get_token() or return;   	# undef if none 
  my $type = $t->[0];                       	# get the type
  $type eq 'T' && $t->[2] and $type = 'TI'; 	# this indicates a comment
  my $val = $t->[1];                        	# get the value
  decode_entities($val);			# convert entities
  $type eq 'S' and $attr = $t->[2];         	# start tag, remember attribute hash for attr()
  $val =~ s/\/$//;                            # suppress trailing '/'
  -t STDIN and print "$type:'$val'\n";        
  "$type:$val";                      		# return type and value 
}

# get start tag attribute
sub attr($)
{
  my $name = shift;					
  return '' unless exists $attr->{$name};       # if exists
  my $value = $attr->{$name};			# get it
  $value =~ s/^\s+|\s+$//g; 			# get rid of extraneous whitespace
  decode_entities($value);			# convert entities
  -t STDIN and print "$name=$value\n";
  $value;
}  

# format and flow text into @lines
# return row and column of first character 
# if contiguous, then the text must not be wrapped
sub flow($$)
{
  my ($text, $contiguous) =@_;

  # get the current last line
  my $last = $lines[$#lines]->[0];		# the current last line
  my $row = $#lines;                  		# remember the original insert point
  my $col = length($last);

  my @l;
  if ($pre)                           		# preformatted
  {
    my $s = "$last$text";	   		# here's the new line
    # but don't let contiguous wrap
    if ($contiguous && $last ne "" && length $s > 79)  
    {
      push @lines, blank;			# force br
      $s = $text;				# and just use the new 
    }
    @l = split "\n",$s,-1;  			# now wrap on '\n's
  } else  
  {
    # not PRE
    $last eq '' and $text =~ s/^\s+//s; 	# clean leading spaces
    $text =~ s/\s+/ /gs;              		# compress intermediates
    
    my $s = "$last$text"; 			# here's the new line

    # don't let contiguous wrap
    if ($contiguous && $last ne "" && length $s > 79) 	
    {
      push @lines, blank;			# force br
      ($s = $text) =~ s/^\s+//s;		# use new text w/o leading spaces
    }

    pos $s = 0;                                 # now wrap it         
    push @l, $1 while $s =~ /\G(.{1,78}$|.{0,78}\s|.{79})\s*/gs; 
    grep { s/^\s+// } @l[1..$#l]                # leading spaces not allowed
  }

  -t STDIN and print "flow $contiguous '$text' ->\n", map { "  '$_'\n" } @l;
      
  if (@l)                                       # did we get anything?
  {
    if (length $l[0] <= length $last)           # did original last line get longer?
    {
      $row++;                                   # no, insert must have wrapped
      $col-=length $l[0];		
    }  
    pop @lines;                                 # drop original last line 
    push @lines, map { [$_, $align] } @l;       # push text and alignment
  }    
  return ($row, $col);                          # return the insert point
}  

sub render($$)
{
  # get html but nuke the baddies
  my $html = pack "C*", grep { $_ == 10 || ($_ >= 32 && $_ < 127) || ($_ >= 160 && $_ <= 255) } unpack "C*", (shift);
  my $istext=shift;

  -t STDIN and print "--------\n$html\n--------\n";
  
  #init
  $title="MBrowse";
  @lines = blank; 
  @items=(); @elements=(); @forms=(); %leftkeys=(); %rightkeys=(); %softkeys=();
  $align = 0; $pre = 0; 
  $baseref="";  

  if ($istext)
  {
    $pre = 1;
    flow ($html, 0);
    return;
  }
 
  $parser = HTML::TokeParser->new(\$html) or die "$!";
  $_ = '';
  $_ = token or return until /^S:html$/;

  my $state = 0;
  for (;;)
  {
    $_ = token or return;
  
    if ($state == 0)
    {
      # outside document
      /^S:head$/ and $state = 1, next;
      /^S:body$/ and $state = 2, next;
    } elsif ($state == 1)
    {
      # head
      /^S:body$/ and do
      {
         push @forms,[base]; # default form action is to reload current page
         $state = 2;
         next;
      };
      
      /^S:title$/ and do
      {
        $_ = token or return until /^T:(.*)$|^E:title$/s;
        $1 and $title = $1;
        next;
      };
      
      # base ref
      /^S:base$/ and do
      {
        my $u = URI->new(attr 'href');
        $u->scheme =~ /^https?$/ and $baseref = $u->canonical; 
        next;
      }; 
      
      # homepage override
      /^S:sethome$/ and do
      {
        # only if current page is trusted
        if ($trusted)
        {
          my $u = URI->new(attr 'href');
          if ($u->scheme =~ /^https?$/) 
          {
            $stack[0]->[0]=$u->canonical; 
            ($home_page = $u->canonical)=~ s/\?.*//;	 # remember home without query
          }
        }    
        next;
      };
      
    } else
    {
      # body
      /^S:br$/ and break, next;

      /^S:p$/ and do
      {
        cbreak; break;
        $_ = attr 'align';
        /left/i and $align=($align & 1)+2, next;  # keep <center> bit intact
        /center/i and $align=($align & 1)+4, next;
        /right/i and $align=($align & 1)+8, next;
        next;
      };
      
      /^E:p$/ and cbreak, $align &= 1, next;
      
      /^S:center$/ and cbreak, $align = 1, next;
      /^E:center$/ and cbreak, $align = 0, next;
      
      /^S:pre$/ and cbreak, $pre=1, next;
      /^E:pre$/ and cbreak, $pre=0, next;

      /^S:a$/ and do 
      {
        
        my $u = URI->new_abs(attr 'href',base);
        $u->scheme =~ /^https?$/ or next;
        {
          $_ = token or return;
          /^T:(.+)/s and do
          {
            push @elements, [0,$u->canonical];
            push @items, [flow(magic.$1,0),\@{$elements[$#elements]}];
            last;
          };  
          /^E:a$/ and last;
          redo;
        }
        next;
      };
      
      /^S:hr$/ and cbreak, flow ('=' x 79,1), next;
   
      # compatibilty hacks     
      /^(S|E):(h\d|div|table|tr|ol|ul|li)$/ and cbreak, next;
      /^S:(td|th)/ and $pre=0, $align&=1, flow(' ',0), next;
       
      # each form is array of [$url, \@element...\@element];
      /^S:form$/ and do
      {
         cbreak;
         my $u = URI->new_abs(attr 'action',base);
         $u->scheme =~ /^https?/ and push @forms, [$u->canonical];
         next;
      };

      /^E:form$/ and cbreak, next;

      # each element is in @element list similar to (1, /@forms name, value (editable), maxlength, size) (see above)
      /^S:input$/ and do
      {
        my $type = attr 'type';
        my $name = attr 'name';
        my $value = attr 'value';
      
        -t STDIN and print " '$type' '$name' '$value'\n";
        $type =~ /^$|text|(pass)/i and do
        {
          $name eq "" and next;
          my $t = (defined $1)?2:1;
          (my $size = attr 'size') =~ s/\D.*//g; 
          $size or $size = 20;       #default size  
          $size > 79 and $size = 79; # limit to 79
          (my $maxlength = attr 'maxlength') =~ s/\D.*//g; $maxlength && $maxlength <= 1024 or $maxlength = 1024; # default length
          
          push @elements, [$t, \@{$forms[$#forms]}, $name, $value, $maxlength, $size]; # add to element list
          push @{$forms[$#forms]}, \@{$elements[$#elements]}; # add it to form
          push @items,[flow("_" x $size,1),\@{$elements[$#elements]}]; # make it selectable
          
          next;
        };  
          
        $type =~ /(checkbox|radio)/i and do
        {
          $name eq "" and next;
          $value eq "" and $value = "on";
          
          my $t = 3;
          my $checked = attr 'checked';
          
          push @elements, [$t, \@{$forms[$#forms]}, $name, $value, $checked]; # add to element list
          push @{$forms[$#forms]}, \@{$elements[$#elements]};  # add it to form
          push @items,[flow("[".nbsp."]",1), \@{$elements[$#elements]}]; # make it selectable if on the page
          next;
        }; 

        $type =~ /hidden/i and do
        {
          $name eq "" and next;
          push @elements, [4, \@{$forms[$#forms]}, $name, $value]; # add to element list
          push @{$forms[$#forms]}, \@{$elements[$#elements]}; # add to form
          next;
        };  

        $type =~ /submit/i and do
        {
          my $bind = attr 'bind'; 
          push @elements, [5, \@{$forms[$#forms]}, $name, $value, $bind]; # add to element list
          push @{$forms[$#forms]}, \@{$elements[$#elements]}; # add to form
          
          if ($bind =~ /sk([1-4])/i)     # softkey?
          {
            $softkeys{$1} = \@{$elements[$#elements]}; # yes, link to softkey list
            next;
          }

          if ($bind =~ /left/i)
          {
            my ($r, $c) = flow($value,1);
            $leftkeys{$r} = \@{$elements[$#elements]}; # link to sidekey list
            next;
          }  
          
          if ($bind =~ /right/i)
          {
            my ($r, $c) = flow($value,1);
            $rightkeys{$r} = \@{$elements[$#elements]}; # link to sidekey list
            next;
          } 
          $value eq "" and $value = "Submit"; 
          push @items,[flow(magic.$value,1),\@{$elements[$#elements]}];
          next;
        };
        next;
      };  

      /^T:(.+)/s and flow($1,0), next;
    }  
  }        
}}

# return fully qualified URL from supplied, current, and $baseref  
sub url($)
{
  my $u = URI->new_abs(shift,base);
  return $u->canonical;
}

# return GET request for specified submit form element. 
sub submit($)
{
  my $e = shift;
  my @f = @{$e->[1]};

  my $base = $baseref || $stack[$#stack]->[0];  
  $base =~ s/\?.*//; 
  my $u = URI->new_abs(shift @f,$base);
  $u->scheme =~ /^https?$/i or $u->scheme("http"); # force http

  my %s;
  $e->[2] and $s{$e->[2]}=$e->[3]; # add our submit
  
  # add form's text, password, checkbox and hidden
  $_->[2] && $_->[0] && $_->[0] <= 4 && ($_->[0]!=3 || $_->[4] ne '') and $s{$_->[2]}=$_->[3] for @f;
  $u->query_form(%s);
  -t STDIN and print "submit $base -> ",$u->canonical,"\n";
  return $u->canonical;
}

# show rendered page to mailbug user
sub show($$)
{
  my $mapsk = 0; # default mappable softkey, 0 if none
  my ($top, $selected)=@_; # get top line and current selected
  $top < 0 || $top > @lines-6 || $selected < 0 || ((@items||$selected) && $selected >= @items) and $top = 0, $selected = 0;


  # first create left margin
  # 0 = left, 1 = center, 2 = p left, 4 = pcenter 8 = pright 

  for (@lines)
  {
    $_->[0] = substr($_->[0],0,79);	     # clip to maximum length	

    my $j = $_->[1];
    if ($j & 8)                              # right justified
    {
      $_->[0] =~ s/^\s+//;
      $_->[1] = 79-(length $_->[0]);
    }  
    elsif ($j & 4 || $j == 1)                # center or <center>
    {
      $_->[0] =~ s/^\s+|\s+$//g;
      $_->[1] = (79-(length $_->[0]))/2;
    } 
    else                                     # left justified
    {
      $_->[0] =~ s/\s+$//;
      $_->[1] = 0;
    }  
  }  


  # get rid of straggling blank lines
  pop @lines while @lines && $lines[$#lines]->[0] !~ /\S/ && !exists $leftkeys{$#lines} && !exists $rightkeys{$#lines};

  # add the left margin to the columns of each item
  $_->[1] += $lines[$_->[0]]->[1] for @items;
 
  if (-t STDIN)
  {
    print "$title\n" if $title;
    for (0..$#lines)
    {
      my $s = $lines[$_]->[0];
      $s =~ s/\x8B/>/g;
      print sprintf("%3d ",$_),("\xB7" x $lines[$_]->[1]),$s,("\xB7" x (79-($lines[$_]->[1]+length($lines[$_]->[0])))),"\n";
    }

    for (@elements)
    {
      print "Element $_: ";
      my $t = $_->[0];
      $t == 0 and print("link '$_->[1]' '",url $_->[1],"'\n"), next; 
      $t == 1 and print("text form=$_->[1] name='$_->[2]' value='$_->[3]' maxlength='$_->[4]' size='$_->[5]'\n"), next;
      $t == 2 and print("password form=$_->[1] name='$_->[2]' value='$_->[3]' maxlength='$_->[4]' size='$_->[5]'\n"), next;
      $t == 3 and print("checkbox form=$_->[1] name='$_->[2]' value='$_->[3]' checked='$_->[4]'\n"), next;
      $t == 4 and print("hidden form=$_->[1] name='$_->[2]' value='$_->[3]'\n"), next;
      $t == 5 and print("submit form=$_->[1] name='$_->[2]' value='$_->[3]'\n  '",submit $_,"'\n"), next;
      print "unknown $t??\n";
    }  
    
    print "Item $_: row=$_->[0] col=$_->[1] element=$_->[2]\n" for @items;

    local $" = " element="; print "Form $_: action=@$_\n" for @forms;

    print "leftkey $_ item=$leftkeys{$_}\n  ",submit $leftkeys{$_},"\n" for keys %leftkeys;
    print "rightkey $_ item=$rightkeys{$_}\n  ",submit $rightkeys{$_},"\n" for keys %rightkeys;
    print "softkey $_ item=$softkeys{$_}\n  ", submit $softkeys{$_},"\n" for keys %softkeys;
  }
  
  lsp::cls;			# clear screen 
  lsp::reset;			# and reset magic switches
  lsp::title $title;		# put up the title
  for (keys %softkeys)		# for each softkey
  {	
    $_>=1 && $_ <= 4 or next;	# just in case
    my $l = $softkeys{$_}->[3]; # get the label
    $l ne "" or $l = "SK$_";    # or default	
    $l =~ s/^!// and $mapsk=$_; # if label starts with '!' we should map it
    lsp::sk $_,$l;		# set it
  }

  @stack>1 and lsp::sk(6,"Back");
  $stack[$#stack]->[0] ne $stack[0]->[0] and lsp::sk(7,"Home");
 
# lsp::setdisable 1;		# always disable when an enabled softkey or sidekey is pressed 
  
  # outer loop, update screen 
  my $oldtop=-1;		# force initial update	
  PAGE: 
  {
    lsp::cursoroff;		# turn cursor off
    lsp::setdisable 31;	        # only let one key queue up	 

    # repaint the page if necessary
    if ($top != $oldtop)
    {
      my ($lks, $rks)=(0,0);    # sidekey bits 
      for (0..5) # for 6 physical lines
      {
        # paint it
        my ($text, $left) = ("",0);
        ($text, $left) = @{$lines[$top+$_]} if $top+$_ < @lines;
        lsp::gotorc $_, $left;	# goto to start
        lsp::clrline;		# clear whole line
        my $ofs = 0;
        while ($text =~ /(.*?)(?:\s{5,}|\s*$)/g) # break out long whitespace
        {
          if ($1 ne '')		# if something to print
          { 
            $ofs and lsp::gotorc $_,$left+$ofs; # move to offset
            lsp::put $1;	# print it
            $ofs = pos $text;	# get next offset
          }
        } 
        # remember what sidekeys should be enabled.
        $lks |= (1<<$_) if exists $leftkeys{$top+$_};
        $rks |= (1<<$_) if exists $rightkeys{$top+$_};
      }  
      lsp::setsidekeys $lks, $rks; # enable the sidekeys
      $oldtop = $top;		# remember top line 
    }
 
    # repaint the items on the page, except selected item 
    for (0..$#items) 
    {
      my ($r, $c, $e) = @{$items[$_]};  # get row, col and \element
      $r >= $top+6 and last;
      ($_ == $selected) || ($r < $top) and next;
      my $t = $e->[0];                  # get type
     
      if ($t == 1 || $t == 2)           # text or password = 1|2, \@form, name, value, maxlength, size
      {                                 
        my $text = substr $e->[3],-($e->[5]);
        $t == 2 and $text =~ s/./*/g;   # password mode, convert to '*'
        lsp::gotorc $r-$top, $c;	# put the cursor
        lsp::put $text; 		# put the text
        next;
      }
    
      if ($t == 3 && $e->[4])           # checkbox = 3, \@form, name, value, checked
      {                                   
        lsp::gotorc($r-$top, $c+1);	# put a '*' if checked
        lsp::put '*'; 			
        next;
      } 
    } 
    lsp::setdisable 1;			# disable on enabled side or softkey

    # collect interesting facts about current item 
    my ($row, $col, $element, $above, $below, $type); 
           
    if (@items)			   
    {
      ($row, $col, $element)=@{$items[$selected]};
      $type = $element->[0]; 
      # find item above and below the current
      $items[$_]->[0] != $row and $above = $_, last for reverse 0..$selected-1;
      $items[$_]->[0] != $row and $below = $_, last for $selected+1..$#items;
    }

    # determine which arrows to enable
    my $enable=1; 					# always enable the side and softkeys
    $enable += 2 if $top || $selected;                  # can go up
    $enable += 4 if ($top<@lines-6) || $selected<$#items;  # can go down
    $enable += 8 if @items;                             # can go left or right
      
    # paging routines bound to keyname returned by getkey
    my %arrows = (
      UP => sub 
      {
        if (@items && $row >= $top)			# if selected not already above top line
        {
          if (defined $above)   			# if an item to go up to...
          { 
            $selected = $above;				# make it current
            $items[$selected]->[0] >= $top and return;	# done if on-screen
          } 
          elsif ($selected)				# or go left
          { 
            $selected--;				# make it current
            $items[$selected]->[0] >= $top and return;	# done if on-screen
          }
        } 
        $top-=6;					# else page up	
        $top < 0 and $top = 0;
        return;
        
      },
   
      LEFT => sub 
      {
        if (@items && $row >= $top && $selected)	# if an item to go left to
        {
          $selected--;					# go	
          $items[$selected]->[0] >= $top and return;	# if if on-screen
        }  
        $top-=6;					# else page up
        $top < 0 and $top = 0;
        return; 
      },  
   
      DOWN => sub
      {
        if (@items && $row < $top+6)
        {
          if (defined $below)				# if an item to go down to
          {
            $selected = $below;				# go	
            $items[$selected]->[0] < $top+6 and return;	# done if on-screen
          }
          elsif ($selected < $#items)
          {
            $selected++;
            $items[$selected]->[0] < $top+6 and return;	# done if on-screen
          }
        }
           
        $top+=6;					# else page down
        $top > @lines-6 and $top = @lines-6;
        $top <= 0 and $top = 0, return;
        $top++ while $top <= @lines-6 && $lines[$top]->[0] eq '';
      },
   
      RIGHT => sub					# ditto for right arrow
      {
        if (@items && $row < $top+6 && $selected < $#items)
        {
          $selected++;					# go to next if possible
          $items[$selected]->[0] < $top+6 and return;	# done if on-screen
        }  
        $top+=6;					# else page in that direction
        $top > @lines-6 and $top = @lines-6;
        $top <= 0 and $top = 0, return;
        $top++ while $top <= @lines-6 && $lines[$top]->[0] eq '';
      }  
    );

    # edit selected item
    if (@items && $row >= $top && $row < $top+6) # if it's on-screen
    { 
      # link or submit
      if ($type == 0 || $type == 5)
      {
        lsp::gotorc $row-$top, $col;      # put cursor
        lsp::select;                      # use the fat right arrow
        lsp::sk 5,"Select";               # show the softkey
        lsp::mapsk 5;			  # map it to enter
        lsp::setdisable 1;                # disable on enabled softkey or sidekey 
        lsp::setkeys $enable;             # enable arrows
        {
          $_ = lsp::getkey();
          /LK([1-6])/ and return submit $leftkeys{$top+$1-1}, $top, $selected;     
          /RK([1-6])/ and return submit $rightkeys{$top+$1-1}, $top, $selected;     
          /SK([1-4])/ and return submit $softkeys{$1}, $top, $selected;
          /SK5/ and return $type?submit $element:url $element->[1], $top, $selected;
          /SK6/ and return undef;
          /SK7/ and return $stack[0]->[0], $top, $selected; 
          exists $arrows{$_} and $arrows{$_}->(), redo PAGE;
          redo;
        }   
      }       
                   
      # text
      if ($type == 1 || $type == 2 )            
      { 
        lsp::sk 5,"";
        lsp::mapsk $mapsk; 
        lsp::setkeys $enable+0x10;          # enable arrows and keyboard
        FIELD: 
        {
          lsp::cursoroff;
          my $text = substr $element->[3],-($element->[5]-1); # room for trailing '_' 
          $type == 2 and $text =~ s/./*/g;  # password shows '*'s
          lsp::gotorc $row-$top,$col;       # put cursor
          lsp::put $text;                   # put the text
          lsp::cursoron;                    # block cursor   
          {
            $_ = lsp::getkey();
  
            exists $arrows{$_} and $arrows{$_}->(), redo PAGE;
            /LF/ and do
            {
              $arrows{RIGHT}->(); 
              redo PAGE;
            };
            /LK([1-6])/ and return submit $leftkeys{$top+$1-1}, $top, $selected;     
            /RK([1-6])/ and return submit $rightkeys{$top+$1-1}, $top, $selected;     
            /SK([1-4])/ and return submit $softkeys{$1}, $top, $selected;
            /SK6/ and return undef;
            /SK7/ and return $stack[0]->[0], $top, $selected;
  
            if (/^.$/)                      # any single character (is printable)	
            {
              length $element->[3] < $element->[4] or redo; # at maxlength
              if ($_ eq ' ' || $_ eq chr 160) # space or nbsp 
              {
                $element->[3] =~ /^$| $/ and redo; # avoid leading or duplicate spaces
                $_ = ' '; # avoid nbsp in text
              }
             
              $element->[3] .= $_;          # append the char
              length $element->[3] > $element->[5]-1 and redo FIELD; # roll left if past max size
              lsp::put $type==1?$_:'*';     # else just put the character
              redo;
            }
    
            if (/BS/)
            {
              $element->[3] =~ /(.*)./ or redo;
              $element->[3] = $1;
              length $element->[3] >= $element->[5]-1 and redo FIELD; # roll right
              lsp::cleft; lsp::put "_"; lsp::cleft;
              redo;
            };
  
            redo;
          }
        } 
      }
          
      # checkbox
      if ($type == 3)             
      {
        lsp::gotorc $row-$top,$col+1;            # put cursor, offset for '['
        lsp::put $element->[4]?'*':' ';          # put star or space  
        lsp::cleft;                              # put cursor on it
        lsp::cursoron;                           # block cursor
        lsp::sk 5,'Toggle';                      # turn on softkey
        lsp::mapsk 5;                     	 # map it to enter
        lsp::setkeys $enable;                    # enable up, down, left, right 
        {
          $_ = lsp::getkey();
          exists $arrows{$_} and $arrows{$_}->(), redo PAGE;
          /LK([1-6])/ and return submit $leftkeys{$top+$1-1}, $top, $selected;     
          /RK([1-6])/ and return submit $rightkeys{$top+$1-1}, $top, $selected;     
          /SK[1-4]/ and return submit $softkeys{$1}, $top, $selected;
          /SK6/ and return undef;
          /SK7/ and return $stack[0]->[0], $top, $selected;
          /SK5/ and do # toggle check
          {
            $element->[4] = !$element->[4];
            lsp::put $element->[4]?'*':' ';
            lsp::cleft;
            lsp::setkeys $enable;                # re-enable after softkey
          };
          redo;
        }   
      }
    }

    # off screen or etc.
    lsp::sk 5,"";				 # softkey off
    lsp::mapsk $mapsk;				 # map default softkey
    lsp::setkeys $enable;			 # enable 
    {
      $_ = lsp::getkey;
      exists $arrows{$_} and $arrows{$_}->(), redo PAGE;
      /LK([1-6])/ and return submit $leftkeys{$top+$1-1}, $top, $selected;     
      /RK([1-6])/ and return submit $rightkeys{$top+$1-1}, $top, $selected;     
      /SK([1-4])/ and return submit $softkeys{$1}, $top, $selected;
      /SK6/ and return undef;
      /SK7/ and return $stack[0]->[0], $top, $selected;
      redo;
    }
  }
}

# encode string as a URL
sub encode($)
{
  my $s = (shift);
  $s=~s/([^-0-9A-Z_.!~*'()])/sprintf "%%%.2X",ord $1/ige;
  $s;
}

# this is the external function
sub browse($$@)
{
  $home_page = shift or die "must specify 'home' url\n";
  my $timeout = (shift)||0; # timeout in seconds
  my $cookie; $cookie.=encode($_)."=".encode(${{@_}}{$_}).";" for keys %{{@_}}; # parse array to "key=value;key=value..." 

  my $ua = LWP::UserAgent->new;			 # create our user agent
  $ua->agent("Lynx/1.0 (compatible; MBrowse 1.0)");			 # user agent

  my $u = URI->new($home_page);
  $u->scheme =~ /^https?$/ or $u->scheme('http'); 
  push @stack, [$u->canonical,0,0];              # save it on the stack
  ($home_page = $u->canonical)=~ s/\?.*//;	 # remember home without query

  $timeout and $timeout+=time;

  {
    $timeout && time > $timeout and return; 	 # done if timeout
       
    lsp::cls; lsp::put "Please wait..."; lsp::twirl;
    my $text = undef; my $type = 0;

    my ($url,$top,$selected)=@{$stack[$#stack]}; # get top of stack

    warn "GET $url\n";
    my $req = HTTP::Request->new(GET => $url); # ok, do the GET

    # is this page trusted?
    (my $here = $url) =~ s/\?.*//;
    my $host = $req->uri->host;
    #warn "HERE=$here HOST=$host\n";
    $trusted = ($here eq $home_page || $host eq 'localhost' || $host eq 'www.landel.com');
    
    # pass cookie only to trusted page  
    $trusted and $cookie and $req->header('Cookie'=>$cookie), warn "GET cookie $cookie\n"; 

    # send request
    my $res = $ua->request($req);
    $url=$stack[$#stack]->[0]=$res->base->canonical; # use the REAL url (canonical or redirected)
    warn "GOT $url\n";
    
    if (!$res->is_success)
    { 
      warn("GET failed\n"),
      $text = "<html><body>
               Sorry, the document is unavailable. The server may be down temporarily. ";
      @stack > 1 and $text.= "Please press 'Back' and try again later.";
      $text .= "</body></html>"; 
    }
    elsif ($res->content_type !~ /^text\/(html|plain)$/)
    { 
      warn("GET content_type:",$res->content_type);
      $text = "<html><body>
               Sorry, the document is type '".$res->content_type."' which can't be displayed.";
      @stack > 1 and $text.= "Please press 'Back' and try another."; 
      $text .= "</body></html>"; 
    } 
    else
    {
      # ok, set text and type 
      $text = $res->content; 
      $type = $res->content_type eq "text/plain";
    }

    render ($text, $type);

    ($url,$top,$selected)=show($top,$selected); # display using stacked context
    if (defined $url)				# get new url 	
    {
      @{$stack[$#stack]}[1,2]=($top,$selected); # remember context for 'back'
      push @stack, [$url,0,0];                  # push onto stack
      redo;					# etc
    }
    # here, go 'back' 
    @stack > 1 or return;			# shouldn't happen
    pop @stack;					# here, pop broken url
    redo;
  }   
}

1;



