#!/usr/bin/perl -w
use warnings;
use strict;

my $debug = 0;
 
my $flashsize=131072; # default 128K flash size
my $flashblock=128;

warn "LSL compiler XX03.07.b (refrigerate after opening)\n";

# max target ram available for variable storage
my $MAXRAM = \16384;                

# max string length
my $MAXLEN = \1023; 

# these hold the compiled output
my $constant_table="";              # table of constant data 
my $constant_index_table="XXX";     # table of 3 byte offsets to constant_table, index zero is not used
my $variable_index_table="XX";      # table of 2 byte offsets to variable RAM, index zero is not used
my $target_index_table="";          # table of 3 byte offsets to code_table
my $code_table="";                  # table of VML code

# each opcode is array of potential forms
# each consisting of opcode value and legal operand list  
# where 'sv' = svar, 'sc' = string, 'svt' = tied svar, 's' = any of these
#       'nv' = nvar, 'nc' = number, 'nvt' = tied nvar, 'n' = any of these
#       't' = target
# leading ? make oeperand optional, if not given compiles a 0 (which is an invalid index into constant data)
my %opcodes = ('set'     => [[0x01, qw(sv ?s)],  [0x02, qw(nv ?n)]],
               'eq'      => [[0x03, qw(s s t)], [0x04, qw(n n t)]],
               'ne'      => [[0x05, qw(s s t)], [0x06, qw(n n t)]],
               'lt'      => [[0x07, qw(s s t)], [0x08, qw(n n t)]],
               'gt'      => [[0x09, qw(s s t)], [0x0A, qw(n n t)]],
               'test'    => [[0x0B, qw(n n t)]],
               'add'     => [[0x0C, qw(nv n)]],
               'sub'     => [[0x0D, qw(nv n)]],
               'mul'     => [[0x0E, qw(nv n)]],
               'div'     => [[0x0F, qw(nv n)]],
               'mod'     => [[0x10, qw(nv n)]],
               'and'     => [[0x11, qw(nv n)]],
               'or'      => [[0x12, qw(nv n)]],
               'xor'     => [[0x13, qw(nv n)]],
               'cat'     => [[0x14, qw(sv s ?n)]],
               'len'     => [[0x15, qw(nv s)]],
               'pos'     => [[0x16, qw(nv s s ?n)]],
               'count'   => [[0x17, qw(nv s ?s)]],
               'pick'    => [[0x18, qw(sv n ?n)]],
               'cut'     => [[0x19, qw(sv n ?n)]],
               'convert' => [[0x1A, qw(sv n ?n)], [0x1B, qw(nv s ?n)]],
               'goto'    => [[0x1C, qw(t)]],
               'call'    => [[0x1D, qw(t)]],
               'return'  => [[0x1E]],
               'tag'     => [[0x1F, qw(sv s)]],
               'untag'   => [[0x20, qw(?sv)]],
               'clear'   => [[0x21]],
               'alias'   => [[0x22, qw(n ?sv)]],
               'store'   => [[0x23, qw(n s)]],
               'fetch'   => [[0x24, qw(s n)]],
               'capture' => [[0x25]],
               'display' => [[0x26, qw(s)]],
               'put'     => [[0x27, qw(s n ?n ?n)]], 
               'glue'    => [[0x28, qw(n ?sv ?s n n ?n ?n ?n ?n ?n ?n)]],
               'wait'    => [[0x29, qw(?n ?n)]],
               'connect' => [[0x2A, qw(s s s ?s)]],
               'reboot' =>  [[0xFF, qw(?n)]]);


# error control 
my $line = 0;
$|=1;
$SIG{__WARN__} = sub { warn @_ if $^S; print $line?"Line $line: @_":@_ };
$SIG{__DIE__} = sub { die @_ if $^S; die $line?"Line $line: @_":@_; };

# all symbols except targets
my %symbol;   # symbol => [type, value] where type is one of sv, nv, svt, nvt, or d                 
sub mksymbol($$$)
{
  my ($name, $type, $value)=@_;
    $name =~ /^[A-Z_]\w*$/i or die "invalid symbol name\n";
  $symbol{$name} and die "duplicate definition";
  $symbol{$name}=[$type, $value];
}  

# create canned tied variables
{ 
  my $index = 0; # remember tied variable index
  # define default tied variables 
  mksymbol $_,"svt",$index++ for qw(VERSION SCRIPTID TITLE 
                                    SAVE0 SAVE1 SAVE2 SAVE3 SAVE4 SAVE5 SAVE6 SAVE7 SAVE8 SAVE9 SAVE10 SAVE11 SAVE12 SAVE13 SAVE14 SAVE15 
                                    SAVE16 SAVE17 SAVE18 SAVE19 SAVE20 SAVE21 SAVE22 SAVE23 SAVE24 SAVE25 SAVE26 SAVE27 SAVE28 SAVE29 SAVE30 SAVE31 
                                    ALIAS0 ALIAS1 ALIAS2 ALIAS3 ALIAS4 ALIAS5 ALIAS6 ALIAS7 ALIAS8 ALIAS9 ALIAS10 ALIAS11 ALIAS12 ALIAS13 ALIAS14 ALIAS15 
                                    ALIAS16 ALIAS17 ALIAS18 ALIAS19 ALIAS20 ALIAS21 ALIAS22 ALIAS23 ALIAS24 ALIAS25 ALIAS26 ALIAS27 ALIAS28 ALIAS29 ALIAS30 ALIAS31 
                                    SERVER 
                                    SOFTKEY1 SOFTKEY2 SOFTKEY3 SOFTKEY4 SOFTKEY5 SOFTKEY6 SOFTKEY7 SOFTKEY8 
                                    CSOFTKEY1 CSOFTKEY2 CSOFTKEY3 CSOFTKEY4 CSOFTKEY5 CSOFTKEY6 CSOFTKEY7 CSOFTKEY8
                                    STAMP);
  mksymbol $_,"nvt",$index++ for qw(FREE RECORDS LED 
                                    TIMER0 TIMER1 TIMER2 TIMER3 TIMER4 TIMER5 TIMER6 TIMER7 
                                    TENTHS0 TENTHS1 TENTHS2 TENTHS3 TENTHS4 TENTHS5 TENTHS6 TENTHS7 
                                    LINE IDLE BACKLIGHT CLOCK SCONFIG SERIAL SPEND SAVAIL CONSOLE CAVAIL RANDOM RESULT WKEY
                                    USED CALLBACK);

  mksymbol $_,"svt",$index++ for qw(SAVE32 SAVE33 SAVE34 SAVE35 SAVE36 SAVE37 SAVE38 SAVE39 SAVE40 SAVE41 SAVE42 SAVE43 SAVE44 SAVE45 SAVE46 SAVE47 
                                    SAVE48 SAVE49 SAVE50 SAVE51 SAVE52 SAVE53 SAVE54 SAVE55 SAVE56 SAVE57 SAVE58 SAVE59 SAVE60 SAVE61 SAVE62 SAVE63); 

  # add new ones to the end!

}

{
  @ARGV or die "Usage: lsl (-s | [-d] [-256] filename)\n";
  $ARGV[0] =~ /^-d(\d*)/i and shift, $debug=$1||1, redo;
  $ARGV[0] =~ /^-256/ and shift, $flashsize=262144, $flashblock=256, redo;
  if ($ARGV[0] =~ /^-s/i)
  {
    # dump .h file for lsl interpreter to stdout and exit
    print "#define MAXRAM $$MAXRAM\n";
    print "#define MAXLEN $$MAXLEN\n";
  
    print "#define $symbol{$_}->[0]_$_ $symbol{$_}->[1]\n" for keys %symbol;
    
    for my $s (keys %opcodes)
    {
      if (@{$opcodes{$s}} == 1) 
      {
        print "#define op_$s $opcodes{$s}->[0][0]\n";
      } else  
      {
        print "#define op_${s}_$_->[1] $_->[0]\n" for @{$opcodes{$s}}
      }  
    }
    exit;
  }
}

# encode number as 24 bit string
sub point($) { substr pack("N", shift), 1 }

# given name, type anmd size, allocate variable and create symbol 
my $alloted = 0;                    # how much allocated so far
sub mkvar($$$)
{
  my ($name, $type, $size)=@_;
  my $index = length($variable_index_table)/2; $index >= 16384 and die "too many variables\n";
  my $ofs = $alloted;
  $variable_index_table .= pack "n", $ofs; # append 16 bit buffer index, offset by two for tag index
  $alloted += $size; $alloted >= $$MAXRAM and die "variable data overflow\n";
  mksymbol $name, $type, $index;
  $debug >= 2 and warn "variable at offset $ofs as index $index \n";
}  

my $scriptid;                       # specified scriptid, for endgame

# build constant and return index
# constant strings and numbers are not referenced via the symbol table
sub mkconst($)                      # build constant entry
{
  my $value=shift;
  my $index = length($constant_index_table)/3; $index >= 32768 and die "too many constants\n";
  my $ofs = length $constant_table;
  $constant_index_table .= point length $constant_table; # append pointer to data
  $constant_table .= $value; length $constant_table > 0xFFFFFF and die "constant table overflow\n";
  $debug >= 2 and warn "constant ", length $value," bytes at offset $ofs as index $index\n"; 
  $index;
}  

my %string;                         # hash of string=>index in constant_index_table
my %number;                         # hash of number=>index in constant_index_table

my %target;                         # symbol=>index in target_index_table
my %forward;                        # symbol=>[offsets in code table]

my $svars=0; my $nvars=0;           # for endgame

# parse numeric expression and return value or undef 
#sub number($)
#{
#  my $t = shift;
#  {
#    local $SIG{__WARN__}=sub{die};          # hide warnings
#    eval "use warnings; \$t=int($t)%2**32"; # do the eval 
#  }  
#  $@?undef:$t;                              # return result
#}

# parse numeric expression
# intermediate results must be less than 2**40
sub number($)
{
  local $SIG{__DIE__};
  local $SIG{__WARN__};
  my $n;

  sub calc($)                      
  {
    $_ = shift;
    sub ev($$$){ my $n = eval "@_"; $@ and die $@; $n }; # eval found parameters with error recover
    1 while s/~\s*([-+]?\d*?\.?\d+\s*)/ev($1,"^","-1")/ex; # handle unary '~'
    for my $op qw (\*\* <<|>> \*|\/|\% \+|- \||\&|^)  # for each operator in order of precedence
    {
      1 while s/(\s*[-+]?\d*?\.?\d+\s*) ($op) (\s*[-+]?\d*?\.?\d+\s*)/"+".(ev($1, $2, $3) % 2**40)/ex;
    }
    /^[+-]?\d*\.?\d+$/ or die;
    $_ % 2**40
  }

  # convert hex and chars to flat decimal
  $_ = shift;
  $debug >= 4 and print "Number in: $_\n";
  1 while (s/(?<!\d)(0x[0-9a-f]+|0b[01]+|0[0-7]+)/oct $1/ie);
  1 while (s/'(.)'/ord $1/e);
  # evaluate subexpressions then the entire result
  # do in an eval to trap divide by zero, etc.
  eval '1 while (s/(?<!\d)\(([^(]*?)\)(?!\d)/calc($1)/ex); $n=calc($_)';
  $n=$@?undef:$n%2**32;
  $debug >= 4 and print "Number out: ",defined($n)?$n:"undef","\n";
  $n;
}

# parse string expression and return result or undef
sub string($)
{
  my $s = shift;
  $debug >= 4 and print "String in: $s\n";
  $s =~ s/^"(.*)"$/$1/s or return undef;
  $s =~ s/([^\\])\^(.)/join "",map { chr $_ } ord $1..ord $2/seg; # expand ranges A^Z, etc
  $s =~ s/\\n/\n/sg;                 # \n => "\n"
  $s =~ s/\\([\\^"])/$1/sg;          # unescape magic chars
  $s =~ tr/\xA0/ /;                 # change nbsp to space
  $s =~ tr/\n\x20-\x7E\xA1-\xFF/ /cd; # delete cruft
  length $s > $$MAXLEN and die "string exceeds $$MAXLEN characters\n"; 
  $debug >= 4 and print "String out: '$s'\n";
  $s;
} 

# parse starts here
# given set of operands, and set of allowable types
# return reference to array of parsed operands or undef
sub parse($$)
{
  my ($args, $expect) = @_;
  my @p;

  $debug >= 3 and warn "expect @$expect\n";
  @$args > @$expect and return undef;                                     # done if too many args      
  my $x;
  
  for (0..$#$expect)                                                      # for each expected arg 
  {
    my $t = $expect->[$_];                                                # get it  
    my $a = $args->[$_];                                                  # and arg
    $t =~ s/^\?// || defined $a or return undef;                          # done unless arg optional or defined
    defined $a or push(@p, undef), next;                                  # push undef if optional arg not specified
    # parse by expected type
    $t eq 't' && $a =~ /^[A-Z_]\w*!?$/i and push(@p,['t',$a]), next;      # target with possible '!'
    # perform define subst unless literal string
    $a !~ /^"/ and $a=~s/\b([A-Z_]\w*)/(exists($symbol{$1})&&$symbol{$1}->[0] eq 'd')?$symbol{$1}->[1]:$1/egi;
    $t eq 'y' && $a =~ /^[A-Z_]\w*$/i and push(@p,['y',$a]), next;        # generic symbol  
    $t =~ /^sc?/ && defined($x = string $a) and push(@p,['sc',$x]), next; # string constant
    $t =~ /^nc?/ && defined($x = number $a) and push(@p,['nc',$x]), next; # numeric constant
    # if expected type is head of actual symbol type then we can use it
    exists $symbol{$a} && $symbol{$a}->[0] =~ /^$t/ and push(@p,[$symbol{$a}->[0],$a]), next;
    return undef;                                                         # sorry
  }
  # all args parsed ok
  $debug >= 3 and warn "parsed ", map(defined $_?"$_->[0]:$_->[1] ":"- ", @p),"\n";
  return \@p;
}  

my $ops = 0;

LINE: while (<>)
{
  chomp;
  $line++;                        # another line
  /^\s*([^#\s]+)\s*/ig or next;   # get the opcode
  my $op = $1;                    # remember the opcode   
  my @args = map { /\S/?$_:undef } /\G\s*(?=[^\#])("(?:\\.|.)*?"|.*?)\s*(?:,|(?=\#)|$)/ig; # parse operands, map empties to undef
  $' =~ /^[^#]/ and die "syntax error before '$''\n"; # croak if anything left besides comment
  
  EXPAND: for my $a (@args)       # convert a 'dot' into a "string" with successive source lines
  {
    $a && $a eq '.' or next;     # find it
    $a = '';                      # wipe it
    my $start = $line;            # remember for error
    while (<>)                    # for each line
    {
      $line++;                    # count it
      s/\s+$//;                   # trim trailing spaces
      $_ eq '.' and do            # if a dot
      {
        $line > $start+7 and warn "warning: text block started in line $start exceeds 6 lines\n";
        $a =~ s/\\n$//;           # kill trailing \\n
        $a = '"'.$a.'"';          # enclose in quotes to make 'string()' happy
        last EXPAND;              # and we are done
      };
      $a .= "$_\\n";              # append to $a with literal \n
    }
    die "no closure for text block started in line $start\n";
  }  
  
  $debug and warn sprintf("%.4X",length $code_table), ": $op", map(defined $_?" $_":" -", @args),"\n";

  # check for various declarations
  $op =~ /^:(.+)$/ and do                                                # :target
  {
    my $name = $1;
    $name =~ /^[A-Z_]\w*!?$/i or die "invalid target\n";                 
    my $forget = $name =~ s/!$//;
    exists $symbol{$name} && $symbol{$name}->[0] ne 't' and die "duplicate symbol\n";
    $forget and delete $symbol{$name};
    my $index = length($target_index_table)/3; $index > 65535 and die "too many targets\n";
    $target_index_table .= substr pack("N", length $code_table),1;
    mksymbol($name,'t',$index);
    if ($forward{$name})
    {
      substr($code_table,$_,2)=pack "n", $index for @{$forward{$name}}; 
      delete $forward{$name};
    }  
    next;
  }; 
    
  $op eq 'id' and do                                                   # id "string"
  {
    $scriptid and die "duplicate id\n";
    my $p = parse(\@args, ['sc']) or die "invalid operand\n"; 
    $scriptid = $p->[0][1];
    next;
  };

  $op eq 'string' and do                                              # string name, size
  {
    my $p = parse(\@args,['y','nc']) or die "invalid operand\n";
    $svars++;
    $p->[1][1] > $$MAXLEN and die "svar length cannot exceed $$MAXLEN\n";
    mkvar $p->[0][1], "sv", $p->[1][1]+1;
    next;
  };

  $op eq 'number' and do                                              # nvar name
  {
    my $p=parse(\@args,['y']) or die "invalid operand\n";
    $nvars++;
    mkvar $p->[0][1], "nv", 4;
    next;
  };

  $op eq 'define' and do                                              # define name, literal
  {
    my $p;
    $p=parse \@args, ['y','n'] or $p=parse \@args, ['y','s'] or die "invalid operand\n";
    $p->[1][0] eq 'sc' and $p->[1][1] = '"'.$p->[1][1].'"';           # requote literal strings
    mksymbol $p->[0][1], 'd', $p->[1][1];                             # add to symbol table as type 'd' 
    next;
  };
  
  # ok, is it an opcode?
  exists $opcodes{$op} or die "invalid instruction\n";

  # yes, could have multiple possible operand lists, find the first that fills the bill
  for (@{$opcodes{$op}})
  {
    my @expect = @$_;
    my $code = shift @expect;
    my $p = parse(\@args, \@expect) or next;                          # can we use it?
    my $ct = length $code_table; 
    $code_table .= chr($code);                                        # yes, compile opcode
    for (@$p)                                                         # for each parsed operand
    {
      $_ or $code_table .= pack("n", 0), next;                        # undef just compiles 0 
      
      my ($t,$v) = @$_;                                               # get type and value  
    
      $t eq 'sc' and do                                               # string
      {
        exists $string{$v} or $string{$v}=mkconst $v."\0";            # compile it if necessary
        $code_table .= pack "n", $string{$v};                         # compile index
        next;
      };  
    
      $t eq 'nc' and do                                               # number
      {
        exists $number{$v} or $number{$v}=mkconst pack "N",$v;        # compile it if necessary
        $code_table .= pack "n", $number{$v};                         # compile index
        next;
      };  
      
      $t =~ /^[sn]v$/                                                 # svars and nvars 
        and $code_table .= pack("n", 0x8000+$symbol{$v}->[1]), next;  # compile as 14 bit index with msbs set to 10
      
      $t =~ /^[sn]vt$/                                                # tied variables
        and $code_table .= pack("n", 0xC000+$symbol{$v}->[1]), next;  # compile as 14 bit index with msbs set to 11
      
      $t eq 't' and do
      {
        my $forget = $v =~ s/!$//;                                    # '!' at the end, delete existing target
        $symbol{$v} and $symbol{$v}->[0] ne 't' and die "not a target\n";  
        $forget and delete $symbol{$v};  
        exists $symbol{$v} or push @{$forward{$v}},length $code_table;
        $code_table .= pack "n", exists $symbol{$v}?$symbol{$v}->[1]:0;
        next;
      };
      
      die "urk, parser created bogus type '$t', value '$v'\n";                      
    } 
    # ok, done
    $ops++;
    $debug > 1 and warn sprintf("CT=%.6X CI=%.6X VI=%.6X TI=%.6X CODE=%.6X",length $constant_table, 
                                                                            length($constant_index_table)/3, 
                                                                            length($variable_index_table)/2, 
                                                                            length($target_index_table)/3, 
                                                                            length $code_table),
                        sprintf(" [%.2X", unpack "C", substr ($code_table, $ct, 1)),                                                                
                        map (sprintf(" %.4X", $_), unpack "n*", substr ($code_table, $ct+1)),"]\n";
    next LINE;
  }
  
  # here, ran out of opcode forms
  die "invalid operand\n";
}  

# endgame
$line = 0;
sub srt($$) { my ($t1, $v1, $t2, $v2) = (@{$symbol{$_[0]}}, @{$symbol{$_[1]}});
              $t1 cmp $t2 
              || ($v1 =~ /\D/ || $v2 =~ /\D/ and return $v1 cmp $v2)
              || $v1 <=> $v2 }
$debug and do
{
  print "$_: $symbol{$_}->[0] $symbol{$_}->[1]\n" for sort srt keys %symbol;
};  

$scriptid or die "must specify script id\n";

# check for unresolved forward referenced
die "unresolved forward references: ", map("$_ ", keys %forward),"\n" if keys %forward;

sub crc($)
{
  my @table = ( 0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241, 0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
                0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40, 0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
                0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40, 0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
                0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641, 0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
                0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240, 0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
                0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41, 0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
                0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41, 0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
                0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640, 0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
                0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240, 0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
                0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41, 0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
                0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41, 0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
                0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640, 0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
                0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241, 0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
                0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40, 0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
                0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40, 0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
                0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641, 0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040 );
   
   my $crc = 0;
   $crc = $table[($crc^(ord $_))&0xff]^($crc>>8) for split //,$_[0]; 
   $crc;
}

# must add a final index entry for getvarmax() function on target
$variable_index_table .= pack "n", $alloted;  

# generate output, note each section has associated 24 bit length
my $output = $scriptid."\n". # preface with single line script id
             point(length $constant_table). $constant_table.
             point(length $constant_index_table). $constant_index_table.  
             point(length $variable_index_table). $variable_index_table.
             point(length $target_index_table). $target_index_table.
             point(length $code_table). $code_table;

my $crc = crc $output;
$output .= pack "n", $crc; # append with crc

print "'$scriptid' is ",length($output)," bytes:\n".  
      "  script id      = ",length($scriptid)+1," bytes,\n". 
      "  data           = ",length($constant_table)," bytes,\n".
      "  constant index = ",length($constant_index_table)," bytes,\n".
      "  variable index = ",length($variable_index_table)," bytes,\n".
      "  target index   = ",length($target_index_table)," bytes,\n".
      "  code           = ",length($code_table)," bytes,\n".
      "  overhead       = 17 bytes (including crc ",(sprintf "%.4X", $crc),")\n"; 

my $size = (length($output)+3+($flashblock-1))&~($flashblock-1); 
print "Script requires $size bytes of flash.\n";

my $remainder = $flashsize-((32*$flashblock)+$size+3);
$remainder < 0 and die "Sorry, flash is too small for this script!\n";
print "$remainder bytes of flash available for data storage.\n";

print "$svars svars and $nvars nvars use $alloted of $$MAXRAM bytes of ram\n"; 

# name lslbin.1.255 has special significance to dbugd backend
open F, ">lslbin.1.255" or die "can't open lslbin.1.255: $?\n";
binmode F; # damn windows
print F $output; close F;
