package lsp;
# convert STDIO to low level lsp
# i.e. strip illegal characters, perform DEL expansion, etc.
# also perform encryption and character mapping

# this should be invoked AFTER $SIG{__WARN__} handling has been installed
# so warnings from this module can be logged appropriately.

use warnings;
use strict;

use IO::Handle;

use POSIX qw(:termios_h strftime);

# connection tracking
my $start = time();
my $active = 0;

# worst CPS, for timeout calculations
my $worstCPS=25;

# handle broken pipe
$SIG{PIPE} = 'IGNORE';

# enable or disable session debug
my $debug;  
sub debug($)
{
  if (shift)
  {
    $debug and return; 
    $debug="/tmp/mbugd.$$";
    if (open DEBUG,">$debug")
    {
      DEBUG->autoflush(1)
    } else
    {
      warn "warning: can't debug: $!\n";
      $debug = undef;  
    }
  } else
  {
    $debug or return;
    close DEBUG;
    unlink $debug;
    $debug=undef;
  }
}

# set of legal characters 
my %control = ( "\x01" => "SOH", "\x04" => "EOT", "\x07" => "BEL", "\x08" => "BS", "\x09" => "TAB", "\x0A" => "LF", "\x0D" => "CR", "\x1B" => "ESC", "\x7F" => "DEL");
my %ascii; $ascii{chr $_}=chr $_ for 32..126;
my %iso8859; $iso8859{chr $_}=chr $_ for 160..255;

# write to debug file with appropriate conversions
sub _debug($@)
{
  $debug or return;
  my $dir = shift;
  print DEBUG $dir, exists $ascii{$_}?$ascii{$_}:exists $control{$_}?$control{$_}:exists $iso8859{$_}?$iso8859{$_}:sprintf "?%d",ord $_ for @_;
}

# these get used a lot
sub SOH() { "\x01" }
sub EOT() { "\x04" }
sub ESC() { "\x1b" }
sub DEL() { "\x7f" }

# first make STDIN raw
unless (-t STDIN)
{
  my $term = POSIX::Termios->new(); 
  $term->getattr(0); 
  $term->setlflag($term->getlflag & ~(ECHO|ICANON)); 
  $term->setoflag($term->getoflag & ~OPOST);
  $term->setcc(VTIME, 0); 
  $term->setcc(VMIN, 1); 
  $term->setattr(0, &POSIX::TCSANOW);
}

$|=1; # also stdout

# low-level receive
# deciphers, decodes and filters
sub get()
{
  my $c;
  {
    $c = getc; defined $c or die "connection dropped\n"; _debug '<',$c;
    if ($c eq DEL)
    {
      $c = getc; defined $c or die "connection dropped\n"; _debug '<',$c;
      $c = chr ord($c)+128;
      exists $iso8859{$c} and last; 
    } else 
    {
      exists $control{$c} || exists $ascii{$c} and last;
    }
    warn sprintf "warning: ignored incoming 0x%.2X\n", ord $c;
    redo;
  }
  return $c;
};  

# low level send
sub put(@)
{
  my $s = join '',@_;

  for my $c (split '', join '', @_)
  {
    $active = time(); # track broken pipes accurately

    if (exists $control{$c} || exists $ascii{$c})
    {
      print $c or die "lsp broken pipe\n";
      _debug '>',$c;
      next;
    }

    if (exists $iso8859{$c})
    {
      $c = chr ord($c)-128;
      print DEL,$c or die "lsp broken pipe\n";
       _debug '>',DEL,$c;
      next;
    }

    if (ord $c >= 139) # yes this is magic, used by mbrowse
    {
      print ESC,"=K" or die "lsp broken pipe\n";
      _debug '>', ESC, "=", "K";
      next;
    } 

    warn sprintf "warning: suppressed outgoing 0x%.2X\n", ord $c;
  } 
  $active = 0; # can't get broken pipe now
};  

# send formatted LSP command
# if command character is lowercase, format data
# strip invalid characters, expand 8 bit characters
sub command($;@)
{
  my ($command, @data)=@_;
  my $s="";
  my $check = ord $command; # checksum includes command char

  put SOH, $command;
  return if $command =~ /[A-Z]/; 
  # here, send command data
  for (@data)	    # for each list	
  {
    for my $c (split '')  # calculate the checksum
    {
      warn(sprintf "warning: dropped outgoing %.2X\n", ord $c), next unless $c eq "\x0A" || exists $ascii{$c} || exists $iso8859{$c};    
      $s .= $c;
      $check += ord $c;
    }
  }
  put sprintf("%.2X",$check&0xff),$s,EOT; # send it
}

# get anticipated response to given command character 
# if lowercase, expected checksummed response string
# abort connection if timeout or invalid character or bad checksum, etc.
sub response($;$)
{
  my ($expect, $timeout) = @_;
  $timeout && $timeout > 30 or $timeout = 30;
  my $s;

  alarm 0;
  $SIG{ALRM} = sub { die "timeout after $timeout seconds waiting for response '$expect'\n" };
  eval {
    alarm $timeout;
    my ($kx, $kk, $c);
    {  
      1 until get eq SOH;
      die (sprintf "expected response '$expect', got 0x%.2X\n", ord $c) unless ($c = get) eq $expect;
      last if $expect =~ /[A-Z]/;
    
      $kx = get . get; 
      die "expected hex, got '$kx' in response '$expect'\n" unless ($kx =~ /[0-9A-F]{2}/i); 
      $kx = hex $kx;
      $kk = ord $expect; # checksum includes response code
      $s = "";
  
      while (($c = get) ne EOT) { $kk += ord $c; $s .= $c; } 
      $kk &= 255;
      die (sprintf "expected checksum 0x%.2X, got 0x%.2X in response '$expect'\n", $kx, $kk) unless $kk == $kx;
    }
  };
  alarm 0;
  $@ and die $@;
  return $s;
}  


# terminal control, see lsp documentation
sub cup() { put ESC, 'A' }
sub cdown() { put ESC, 'B' }
sub cright() { put ESC, 'C' }
sub cleft() { put ESC, 'D' }
sub cls() { put ESC, 'E' }
sub home() { put ESC, 'H' }
sub scrolldn() { put ESC, 'I' } # move cursor up, scroll down if top of screen
sub clreos() { put ESC, 'J' }
sub clreol() { put ESC, 'K' }
sub insline() { put ESC, 'L' }
sub delline() { put ESC, 'M' }
sub gotorc($$) { my ($row, $col) = @_; put ESC, 'Y', chr $row+32, chr $col+32 }
sub clrsos() { put ESC, 'd' }
sub cursoron() { put ESC, 'e' }
sub cursoroff() { put ESC, 'f' }
sub scrollup() { put ESC, 'i' } # move cursor down, scroll up if bottom of screen  
sub save() { put ESC, 'j' }
sub restore() { put ESC, 'k' }
sub clrline() { put ESC, 'l' }
sub wrapon() { put ESC, 'v' }
sub wrapoff() { put ESC, 'w' }

sub reset() { put ESC, '%' }
sub resetsk { put ESC, '0' }
sub sk($$) { my ($sk, $label)=@_; put ESC, $sk, substr($label,0,8), EOT }
sub mapsk($) { my $sk = shift; put ESC, '#', $sk }
sub title($) { my $s = shift; put ESC, '!',$s,EOT }
sub twirl() { put ESC, '.' };
sub select() { put ESC, '>' };

sub tping() { put ESC,'?' }
sub setmisc($) { my $n = shift; put ESC, '{', chr $n+0x20 } 
sub setkeys($) { my $n = shift; put ESC, '*', chr $n+0x20 } 
sub setsidekeys($$) { my ($left, $right) = @_; put ESC, '@', chr $left+0x20, chr $right+0x20 } 
sub setdisable($) { my $n = shift; put ESC, '$', chr $n+0x20 }

# read key
# return a single character or one of these: 
my %escapes = ( A=>'UP', B=>'DOWN', C=>'RIGHT', D=>'LEFT',
                1=>'SK1', 2=>'SK2', 3=>'SK3', 4=>'SK4', 5=>'SK5', 6=>'SK6', 7=>'SK7',
                '$'=>'DISABLED', '?'=>'PONG',
                l1=>'LK1', l2=>'LK2', l3=>'LK3', l4=>'LK4', l5=>'LK5', l6=>'LK6',
                r1=>'RK1', r2=>'RK2', r3=>'RK3', r4=>'RK4', r5=>'RK5', r6=>'RK6' );

# get a character or control string
sub getkey()
{
  my $s;
  alarm 0;
  $SIG{ALRM} = sub { die "getkey timeout\n" };
  eval
  {
    alarm 120;
    {
      $s = get;		   # get a character
      if ($s ne ESC)
      {
        $s = $control{$s} if exists $control{$s};
        last;
      }  	
      $s = get;                      # get another
      $s =~ /[lr]/ and $s .= get;    # and another if l or r
      redo unless exists $escapes{$s};
      $s = $escapes{$s};
    }
  };
  alarm 0;
  $@ and die $@;
  return $s;	
}    

# lsp commands

# ping, return number of MNP errors  
sub ping()			
{
  command 'P';			# ping	
  return hex response 'p';	# return mnp errors	
}

# return configuration lines as a list
sub config()
{
  command 'C';
  my $config = response 'c';
  my @c = unpack "C*", $config;
  #warn "@c\n";
  split /\n/, $config,-1; # retain trailing fields
}

# return file 0 to 255
sub upload($)
{
  my $file = shift;
  command 'r',sprintf "%X",$file;
  return response 'r', 16384/$worstCPS; # allow the longest possible file
}

# return available free space to store file f
sub free($)
{
  my $file = shift;
  command 'l',sprintf "%X",$file;
  return hex response 'l';
}

# download a file to file handle f
# hopefully there is enough 'free' to hold it   
sub download($$)
{
  my ($file, $data)=@_;
  command 'w',sprintf "%X\n%s", $file, $data;
  response 'W', (length $data) / $worstCPS;
}

# set clock with string in format MMDDYYYYhhmmss+0000 (ZZZ)
sub setclock($)
{
  command 't',shift; 
  response 'T'; 
} 
 
# get statistics record as a list
sub statistics()
{
  command "S";
  return map { hex } split /\n/, response "s";
}

# wait n seconds for STDIN to close
sub hangup($)
{
  my $timeout = shift;
  $timeout && $timeout > 3 or $timeout = 3;

  alarm 0;
  $SIG{ALRM} = sub { die "hangup timeout\n" };
  eval {
    alarm $timeout;
    1 while getc;
  };
  alarm 0;
}

# abort connection with error 
# returns on pipe close or 3 seconds whichever comes first
sub abort($;$)
{
  my $msg = (shift)||"";                # get message
  my $callback = (shift)||0;
  command "e",(sprintf "%X", $callback)."\n$msg"; # send it
  hangup 3+((length $msg)/$worstCPS);		
}

# same for goodbye
sub goodbye($$)
{
  my $msg = (shift) || "";		# get message
  my $callback = (shift)||0;
  command "g",(sprintf "%X", $callback)."\n$msg" ; # send it
  hangup 3+((length $msg)/$worstCPS);
  debug 0;				# nuke debug capture
}

# return connection start and stop times  
sub span()
{
  return $start, $active || time();
}

1;
